#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import traceback
import datetime
import time

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import banner
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import executeRoutes
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.varios import getDateTime

def consolidateResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        citiesIds = getIdsOfCities(session)
        if len(citiesIds) == 0:
            event.set()
            return

        if len(citiesIds) == 1:
            print('You need at least two cities to consolidate resources.')
            event.set()
            return

        banner()
        source_msg = 'Select source cities to send resources from:'
        sourceCities, sourceCitiesDict = ignoreCities(session, msg=source_msg)
        
        if sourceCities is None:
            event.set()
            return

        banner()
        print('Select destination cities to receive resources:')
        destinationCity = chooseCity(session)

        if destinationCity is None:
            event.set()
            return
        
        destination_id_str = str(destinationCity['id'])
        if destination_id_str in sourceCities:
            sourceCities.remove(destination_id_str)
            del sourceCitiesDict[destination_id_str]

        banner()
        print('Define resource limits to keep:')
    
        limits = []
        
        for i, resource in enumerate(materials_names):
            prompt = 'Enter maximum {} to keep (skip to keep everything): '.format(resource)
            resourceLimit = read(msg=prompt, min=-1, default=-1)
            limits.append(resourceLimit)
        
        banner()
        print('Define how often to execute the process in hours (min - 1 hour):')
        intervalInHours = read(min=1, default=1)

        banner()
        print(('The process will transfer everything above specified limits:'))
        for i, resource in enumerate(materials_names):
            if limits[i] == -1:
                print(('- Keep all {}').format(resource))
            else:
                print(('- Keep up to {} {} and send excess').format(addThousandSeparator(limits[i]), resource))
        
        source_city_names_str = ', '.join([city['name'] for city in sourceCitiesDict.values()])
        print(('\nFrom {} to {} every {} hours').format(
            source_city_names_str, 
            destinationCity['name'], 
            intervalInHours,
        ))

        print("\nProceed? [Y/n]")
        rta = read(values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            event.set()
            return

        enter()

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()  # this is where we give back control to main process

    info = 'Consolidate resources from {} to {} every {:d} hours'.format(source_city_names_str, destinationCity['name'], intervalInHours)
    setInfoSignal(session, info)

    nextExecutionTime = datetime.datetime.now() + datetime.timedelta(hours=intervalInHours)

    session.setStatus(
        f" {source_city_names_str} -> {destinationCity['name']} | Next at {getDateTime(nextExecutionTime.timestamp())}"
    )

    try:
        do_it(session, limits, sourceCities, destinationCity['id'], intervalInHours)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session, limits, sourceCityIds, destinationCityId, intervalInHours):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    limits : array
    sourceCityIds : list[str]
    destinationCityId : int
    intervalInHours : int
    """

    firstRun = True
    nextRunTime = datetime.datetime.now()
    total_amount_sent = 0

    while True:
        currentTime = datetime.datetime.now()
        loop_amount_sent = 0
        if currentTime < nextRunTime and not firstRun:
            time.sleep(60)
            continue
            
        # Loop through each source city ID in the provided list
        for sourceCityId in sourceCityIds:
            try:
                sourceCity = getCity(session.get(city_url + str(sourceCityId)))
                destinationCity = getCity(session.get(city_url + str(destinationCityId)))
            except Exception as e:
                print(f"Failed to get data for source city ID {sourceCityId}: {e}")
                continue # Skip to the next city if one fails

            toSend = [0] * len(materials_names)
            totalToSend = 0

            session.setStatus(
                f"{sourceCity['name']} -> {destinationCity['name']}| Processing..."
            )

            for i, resource in enumerate(materials_names):
                limit = limits[i]
                
                sourceAmount = sourceCity['availableResources'][i]
                destinationSpace = destinationCity['freeSpaceForResources'][i]

                if limit == -1 or limit < 0:  # Handle any negative number as "keep all"
                    toSend[i] = 0
                else:
                    excess = max(0, sourceAmount - limit)
                    sendable = min(excess, destinationSpace)
                    toSend[i] = sendable
                    totalToSend += sendable

            if totalToSend != 0:
                route = (
                    sourceCity,
                    destinationCity,
                    destinationCity["islandId"],
                    *toSend,
                )

                executeRoutes(session, [route], useFreighters=False)
                loop_amount_sent += totalToSend
                total_amount_sent += totalToSend

        nextRunTime = datetime.datetime.now() + datetime.timedelta(hours=intervalInHours)
        
        # Get the name of each source city by its ID for the final status message
        source_city_names = [getCity(session.get(city_url + str(city_id)))['name'] for city_id in sourceCityIds]
        source_city_names_str = ', '.join(source_city_names)
        # Get the destination city's name for the status message
        destinationCityName = getCity(session.get(city_url + str(destinationCityId)))['name']

        session.setStatus(
            f"Sent {addThousandSeparator(loop_amount_sent)} resources from {source_city_names_str} -> {destinationCity['name']} | Total sent: {addThousandSeparator(total_amount_sent)} | Next shipment at {getDateTime(nextRunTime.timestamp())}"
        )

        firstRun = False

        time.sleep(60 * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import sys

import requests

from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read



def importExportCookie(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    banner()
    try:
        print("Do you want to import or export the cookie?")
        print("(0) Exit")
        print("(1) Import")
        print("(2) Export")
        action = read(min=0, max=2)
        if action == 1:
            importCookie(session)
        elif action == 2:
            exportCookie(session)

        event.set()
    except KeyboardInterrupt:
        event.set()
        return


def importCookie(session):
    banner()
    print(
        "{}⚠️ INSERTING AN INVALID COOKIE WILL LOG YOU OUT OF YOUR OTHER SESSIONS ⚠️{}\n\n".format(
            bcolors.WARNING, bcolors.ENDC
        )
    )
    print("Go ahead and export the cookie from another ikabot instance now and then")
    print('type your "ikariam" cookie below:')
    newcookie = read()
    newcookie = newcookie.strip()
    newcookie = newcookie.replace("ikariam=", "")
    cookies = session.getSessionData()["cookies"]
    cookies["ikariam"] = newcookie
    if session.host in session.s.cookies._cookies:
        session.s.cookies.set("ikariam", newcookie, domain=session.host, path="/")
    else:
        session.s.cookies.set("ikariam", newcookie, domain="", path="/")

    html = session.s.get(session.urlBase).text

    if session.isExpired(html):
        print(
            "{}Failure!{} All your other sessions have just been invalidated!".format(
                bcolors.RED, bcolors.ENDC
            )
        )
        enter()
    else:
        print(
            "{}Success!{} This ikabot session will now use the cookie you provided".format(
                bcolors.GREEN, bcolors.ENDC
            )
        )
        sessionData = session.getSessionData()
        sessionData["cookies"]["ikariam"] = newcookie
        session.setSessionData(sessionData)
        enter()
    session.get()


def exportCookie(session):
    banner()
    session.get()  # get valid cookie in case user has logged the bot out before running this feature
    ikariam = session.getSessionData()["cookies"]["ikariam"]
    print(
        "Use this cookie to synchronise two ikabot instances on 2 different machines\n\n"
    )
    print("ikariam=" + ikariam + "\n\n")

    cookie = json.dumps(
        {"ikariam": ikariam}
    )  # get ikariam cookie, only this cookie is invalidated when the bot logs the user out.
    cookies_js = 'cookies={};i=0;for(let cookie in cookies){{document.cookie=Object.keys(cookies)[i]+"="+cookies[cookie];i++}}'.format(
        cookie
    )
    print(
        """To prevent ikabot from logging you out while playing Ikariam do the following:
    1. Be on the "Your session has expired" screen
    2. Open Chrome javascript console by pressing CTRL + SHIFT + J
    3. Copy the text below, paste it into the console and press enter
    4. Press F5
    """
    )
    print(cookies_js)
    enter()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import sys
import time
import traceback

import requests

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import run, set_child_mode
from ikabot.helpers.varios import timeStringToSec, wait
from ikabot.helpers.apiComm import getPiratesCaptchaSolution



def autoPirate(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    banner()
    try:        

        print(
            "{}⚠️ USING THIS FEATURE WILL EXPOSE YOUR IP ADDRESS TO A THIRD PARTY FOR CAPTCHA SOLVING ⚠️{}\n\n".format(
                bcolors.WARNING, bcolors.ENDC
            )
        )
        print("How many pirate missions should I do? (min = 1)")
        pirateCount = read(min=1, digit=True)
        print("Should I schedule pirate missions by the time of day? (y/N)")
        scheduleInput = read(values=["y", "Y", "n", "N", ""])
        if scheduleInput.lower() == "y":
            pirateSchedule = True
            print(
                """Which pirate mission should I do at daytime? (Default mission)
        (1) 2m 30s
        (2) 7m 30s
        (3) 15m
        (4) 30m
        (5) 1h
        (6) 2h
        (7) 4h
        (8) 8h
        (9) 16h
        """
            )
            pirateMissionDayChoice = read(min=1, max=9, digit=True)
            print(
                """At which hours should I operate at daytime? (Default: 9 hours from 10 till 18)
            """
            )
            print("From: ")
            dayStart = read()
            if dayStart == "":
                dayStart = 10
            else:
                dayStart = int(dayStart)

            print("Till: ")
            dayEnd = read()
            if dayEnd == "":
                dayEnd = 18
            else:
                dayEnd = int(dayEnd)

            print(
                """Which pirate mission should I do at night time?
            (1) 2m 30s
            (2) 7m 30s
            (3) 15m
            (4) 30m
            (5) 1h
            (6) 2h
            (7) 4h
            (8) 8h
            (9) 16h
            """
            )
            pirateMissionNightChoice = read(min=1, max=9, digit=True)

            print(
                """At which hours should I operate at night time? (Default: 15 hours from 19 till 9): 
            """
            )
            print("From: ")
            nightStart = read()
            if nightStart == "":
                nightStart = 19
            else:
                nightStart = int(nightStart)

            print("Till: ")
            nightEnd = read()
            if nightEnd == "":
                nightEnd = 9
            else:
                nightEnd = int(nightEnd)
        else:
            pirateSchedule = False
            print(
                """Which pirate mission should I do?
        (1) 2m 30s
        (2) 7m 30s
        (3) 15m
        (4) 30m
        (5) 1h
        (6) 2h
        (7) 4h
        (8) 8h
        (9) 16h
        """
            )
            pirateMissionChoice = read(min=1, max=9, digit=True)
        if pirateSchedule == True:
            current_hour = int(time.strftime("%H"))
            if current_hour >= dayStart and current_hour <= dayEnd:
                pirateMissionChoice = pirateMissionDayChoice
            elif current_hour >= nightStart or current_hour <= nightEnd:
                pirateMissionChoice = pirateMissionNightChoice
            else:
                pirateMissionChoice = pirateMissionDayChoice
        print(
            "Do you want me to automatically convert capture points to crew strength? (Y|N)"
        )
        autoConvert = read(values=["y", "Y", "n", "N"])
        if autoConvert.lower() == "y":
            print(
                'How many points should I convert every time I do a mission? (Type "all" to convert all points at once)'
            )
            convertPerMission = read(min=0, additionalValues=["all"], digit=True)
        print(
            "Enter a maximum additional random waiting time between missions in seconds. (min = 0)"
        )
        maxRandomWaitingTime = read(min=0, digit=True)
        piracyCities = getPiracyCities(session, pirateMissionChoice)
        if piracyCities == []:
            print(
                "You do not have any city with a pirate fortress capable of executing this mission!"
            )
            enter()
            event.set()
            return

        print(
            "YAAAAAR!"
        )  # get data for options such as auto-convert to crew strength, time intervals, number of piracy attempts... ^^
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    try:
        while pirateCount > 0:
            session.setStatus("Pirating for " + str(pirateCount) + " more runs")
            if pirateSchedule == True:
                current_hour = int(time.strftime("%H"))
                if current_hour >= dayStart and current_hour <= dayEnd:
                    pirateMissionChoice = pirateMissionDayChoice
                elif current_hour >= nightStart or current_hour <= nightEnd:
                    pirateMissionChoice = pirateMissionNightChoice
                else:
                    pirateMissionChoice = pirateMissionDayChoice
            pirateCount -= 1
            piracyCities = getPiracyCities(
                session, pirateMissionChoice
            )  # this is done again inside the loop in case the user destroys / creates another pirate fortress while this module is running
            if piracyCities == []:
                raise Exception(
                    "No city with pirate fortress capable of executing selected mission"
                )
            html = session.post(
                city_url + str(piracyCities[0]["id"])
            )  # this is needed because for some reason you need to look at the town where you are sending a request from in the line below, before you send that request
            if (
                '"showPirateFortressShip":0' in html
            ):  # this is in case the user has manually run a capture run, in that case, there is no need to wait 150secs instead we can check every 5
                url = "view=pirateFortress&cityId={}&position=17&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1".format(
                    piracyCities[0]["id"], piracyCities[0]["id"], actionRequest
                )
                html = session.post(url)
                wait(getCurrentMissionWaitingTime(html), maxRandomWaitingTime)
                pirateCount += 1  # don't count this as an iteration of the loop
                continue

            url = "action=PiracyScreen&function=capture&buildingLevel={0}&view=pirateFortress&cityId={1}&position=17&activeTab=tabBootyQuest&backgroundView=city&currentCityId={1}&templateView=pirateFortress&actionRequest={2}&ajax=1".format(
                piracyMissionToBuildingLevel[pirateMissionChoice],
                piracyCities[0]["id"],
                actionRequest,
            )
            html = session.post(url)

            if "function=createCaptcha" in html:
                try:
                    for i in range(20):
                        session.setStatus("Resolving captcha " + str(i) + "/20")
                        if i == 19:
                            msg = "Failed to resolve captcha too many times, autoPirate has been terminated."
                            sendToBot(session, msg)
                            raise Exception("Failed to resolve captcha too many times")
                        picture = session.get(
                            "action=Options&function=createCaptcha", fullResponse=True
                        ).content
                        captcha = resolveCaptcha(session, picture)
                        session.setStatus("Got captcha: " + captcha)
                        if captcha == "Error":
                            time.sleep(5)
                            continue
                        session.post(city_url + str(piracyCities[0]["id"]))
                        params = {
                            "action": "PiracyScreen",
                            "function": "capture",
                            "cityId": piracyCities[0]["id"],
                            "position": "17",
                            "captchaNeeded": "1",
                            "buildingLevel": str(
                                piracyMissionToBuildingLevel[pirateMissionChoice]
                            ),
                            "captcha": captcha,
                            "activeTab": "tabBootyQuest",
                            "backgroundView": "city",
                            "currentCityId": piracyCities[0]["id"],
                            "templateView": "pirateFortress",
                            "actionRequest": actionRequest,
                            "ajax": "1",
                        }
                        html = session.post(params=params, noIndex=True)
                        if (
                            '"showPirateFortressShip":1' in html
                        ):  # if this is true, then the crew is still in the town, that means that the request didn't succeed
                            time.sleep(5)
                            continue
                        break
                except Exception:
                    info = ""
                    msg = "Error in:\n{}\nCause:\n{}".format(
                        info, traceback.format_exc()
                    )
                    sendToBot(session, msg)
                    break
            if autoConvert.lower() == "y":
                convertCapturePoints(session, piracyCities, convertPerMission)
            wait(piracyMissionWaitingTime[pirateMissionChoice], maxRandomWaitingTime)

    except Exception:
        info = ""
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
        event.set()
        return


def resolveCaptcha(session, picture):
    session_data = session.getSessionData()
    if (
        "decaptcha" not in session_data
        or session_data["decaptcha"]["name"] == "default"
    ):
        captcha = getPiratesCaptchaSolution(session, picture)
        return captcha
    elif session_data["decaptcha"]["name"] == "custom":
        files = {"upload_file": picture}
        captcha = requests.post(
            "{0}".format(session_data["decaptcha"]["endpoint"]), files=files
        ).text
        return captcha
    elif session_data["decaptcha"]["name"] == "9kw.eu":
        credits = requests.get(
            "https://www.9kw.eu/index.cgi?action=usercaptchaguthaben&apikey={}".format(
                session_data["decaptcha"]["relevant_data"]["apiKey"]
            )
        ).text
        if int(credits) < 10:
            raise Exception("You do not have enough 9kw.eu credits!")
        captcha_id = requests.post(
            "https://www.9kw.eu/index.cgi?action=usercaptchaupload&apikey={}".format(
                session_data["decaptcha"]["relevant_data"]["apiKey"]
            ),
            headers={"Content-Type": "multipart/form-data"},
            files={"file-upload-01": picture},
        ).text
        while True:
            captcha_result = requests.get(
                "https://www.9kw.eu/index.cgi?action=usercaptchacorrectdata&id={}&apikey={}".format(
                    captcha_id, session_data["decaptcha"]["relevant_data"]["apiKey"]
                )
            ).text
            if captcha_result != "":
                return captcha_result.upper()
            wait(5)
    elif session_data["decaptcha"]["name"] == "telegram":
        sendToBot(session, "Please solve the captcha", Photo=picture)
        captcha_time = time.time()
        while True:
            response = getUserResponse(session, fullResponse=True)
            if len(response) == 0:
                time.sleep(5)
                continue
            response = response[-1]
            if response["date"] > captcha_time:
                return response["text"]
            time.sleep(5)


def getPiracyCities(session, pirateMissionChoice):
    """Gets all user's cities which have a pirate fortress in them
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    piracyCities : list[dict]
    """
    cities_ids = getIdsOfCities(session)[0]
    piracyCities = []
    for city_id in cities_ids:
        html = session.get(city_url + city_id)
        city = getCity(html)
        for pos, building in enumerate(city["position"]):
            if (
                building["building"] == "pirateFortress"
                and building["level"]
                >= piracyMissionToBuildingLevel[pirateMissionChoice]
            ):
                piracyCities.append(city)
                break
    return piracyCities


def convertCapturePoints(session, piracyCities, convertPerMission):
    """Converts all the users capture points into crew strength
    Parameters
    ----------
    session : ikabot.web.session.Session
    piracyCities: a list containing all cities which have a pirate fortress
    """
    params = {
        "view": "pirateFortress",
        "activeTab": "tabCrew",
        "cityId": piracyCities[0]["id"],
        "position": 17,
        "backgroundView": "city",
        "currentCityId": piracyCities[0]["id"],
        "templateView": "pirateFortress",
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    html = session.post(params=params)
    rta = re.search(r'\\"capturePoints\\":\\"(\d+)\\"', html)
    capturePoints = int(rta.group(1))
    if convertPerMission == "all":
        convertPerMission = capturePoints
    if "conversionProgressBar" in html:  # if a conversion is still in progress
        return
    data = {
        "action": "PiracyScreen",
        "function": "convert",
        "view": "pirateFortress",
        "cityId": piracyCities[0]["id"],
        "islandId": piracyCities[0]["islandId"],
        "activeTab": "tabCrew",
        "crewPoints": str(int(convertPerMission / 10)),
        "position": "17",
        "backgroundView": "city",
        "currentCityId": piracyCities[0]["id"],
        "templateView": "pirateFortress",
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    html = session.post(params=data, noIndex=True)


def getCurrentMissionWaitingTime(html):
    try:
        match = re.search(r'ongoingMissionTimeRemaining\\":(\d+),', html)
        assert (
            match
        ), "Couldn't find remaining ongoing mission time, did you run a pirate mission manually?"
        return int(match.group(1))
    except: # In case we can not find current mission waiting time, just sleep for 10 mins (Not a perfect solution)
        return 10 * 60



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *


def obtainMiraclesAvailable(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    islands: list[dict]
    """
    idsIslands = getIslandsIds(session)
    islands = []
    for idIsland in idsIslands:
        html = session.get(island_url + idIsland)
        island = getIsland(html)
        island["activable"] = False
        islands.append(island)

    ids, cities = getIdsOfCities(session)
    for city_id in cities:
        city = cities[city_id]
        # get the wonder for this city
        wonder = [
            island["wonder"]
            for island in islands
            if city["coords"] == "[{}:{}] ".format(island["x"], island["y"])
        ][0]
        # if the wonder is not new, continue
        if wonder in [island["wonder"] for island in islands if island["activable"]]:
            continue

        html = session.get(city_url + str(city["id"]))
        city = getCity(html)

        # make sure that the city has a temple
        for i in range(len(city["position"])):
            if city["position"][i]["building"] == "temple":
                city["pos"] = str(i)
                break
        else:
            continue

        # get wonder information
        params = {
            "view": "temple",
            "cityId": city["id"],
            "position": city["pos"],
            "backgroundView": "city",
            "currentCityId": city["id"],
            "actionRequest": actionRequest,
            "ajax": "1",
        }
        data = session.post(params=params)
        data = json.loads(data, strict=False)
        html = data[1][1][1]
        match = re.search(r'<div id="wonderLevelDisplay"[^>]*>\\n\s*(\d+)\s*</div>', html)
        level = 0
        if match:
            level = int(match.group(1))

        data = data[2][1]
        available = data["js_WonderViewButton"]["buttonState"] == "enabled"
        if available is False:
            for elem in data:
                if "countdown" in data[elem]:
                    enddate = data[elem]["countdown"]["enddate"]
                    currentdate = data[elem]["countdown"]["currentdate"]
                    break

        # set the information on the island which wonder we can activate
        for island in islands:
            if island["id"] == city["islandId"]:
                island["activable"] = True
                island["ciudad"] = city
                island["wonderActivationLevel"] = level
                island["available"] = available
                if available is False:
                    island["available_in"] = int(float(enddate)) - int(float(currentdate))
                break

    # only return island which wonder we can activate
    return [island for island in islands if island["activable"]]


def activateMiracleHttpCall(session, island):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict

    Returns
    -------
    json : dict
    """
    params = {
        "action": "CityScreen",
        "cityId": island["ciudad"]["id"],
        "function": "activateWonder",
        "position": island["ciudad"]["pos"],
        "backgroundView": "city",
        "currentCityId": island["ciudad"]["id"],
        "templateView": "temple",
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    response = session.post(params=params)
    return json.loads(response, strict=False)


def chooseIsland(islands):
    """
    Parameters
    ----------
    islands : list[dict]

    Returns
    -------
    island : dict
    """
    print("Which miracle do you want to activate?")
    # Sort islands by name
    sorted_islands = sorted(islands, key=lambda x: x["wonderName"])
    i = 0
    print("(0) Exit")
    for island in sorted_islands:
        i += 1
        if island["available"]:
            print("({:d}) {}".format(i, island["wonderName"]))
        else:
            print(
                "({:d}) {} (available in: {})".format(
                    i, island["wonderName"], daysHoursMinutes(island["available_in"])
                )
            )

    index = read(min=0, max=i)
    if index == 0:
        return None
    island = sorted_islands[index - 1]
    return island


def activateMiracle(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        islands = obtainMiraclesAvailable(session)
        if islands == []:
            print("There are no miracles available.")
            enter()
            event.set()
            return

        island = chooseIsland(islands)
        if island is None:
            event.set()
            return

        if island["available"]:
            print("\nThe miracle {} will be activated".format(island["wonderName"]))
            print("Proceed? [Y/n]")
            activate_miracle_input = read(values=["y", "Y", "n", "N", ""])
            if activate_miracle_input.lower() == "n":
                event.set()
                return

            miracle_activation_result = activateMiracleHttpCall(session, island)

            if miracle_activation_result[1][1][0] == "error":
                print(
                    "The miracle {} could not be activated.".format(
                        island["wonderName"]
                    )
                )
                enter()
                event.set()
                return

            data = miracle_activation_result[2][1]
            for elem in data:
                if "countdown" in data[elem]:
                    enddate = data[elem]["countdown"]["enddate"]
                    currentdate = data[elem]["countdown"]["currentdate"]
                    break
            wait_time = int(float(enddate)) - int(float(currentdate))

            print("The miracle {} was activated.".format(island["wonderName"]))
            enter()
            banner()

            while True:
                print("Do you wish to activate it again when it is finished? [y/N]")

                reactivate_again_input = read(values=["y", "Y", "n", "N", ""])
                if reactivate_again_input.lower() != "y":
                    event.set()
                    return

                iterations = read(msg="How many times?: ", digit=True, min=0)

                if iterations == 0:
                    event.set()
                    return

                duration = wait_time * iterations

                print("It will finish in:{}".format(daysHoursMinutes(duration)))

                print("Proceed? [Y/n]")
                reactivate_again_input = read(values=["y", "Y", "n", "N", ""])
                if reactivate_again_input.lower() == "n":
                    banner()
                    continue
                break
        else:
            print(
                "\nThe miracle {} will be activated in {}".format(
                    island["wonderName"], daysHoursMinutes(island["available_in"])
                )
            )
            print("Proceed? [Y/n]")
            user_confirm = read(values=["y", "Y", "n", "N", ""])
            if user_confirm.lower() == "n":
                event.set()
                return
            wait_time = island["available_in"]
            iterations = 1

            print("\nThe mirable will be activated.")
            enter()
            banner()

            while True:
                print("Do you wish to activate it again when it is finished? [y/N]")

                reactivate_again_input = read(values=["y", "Y", "n", "N", ""])
                again = reactivate_again_input.lower() == "y"
                if again is True:
                    try:
                        iterations = read(msg="How many times?: ", digit=True, min=0)
                    except KeyboardInterrupt:
                        iterations = 1
                        break

                    if iterations == 0:
                        iterations = 1
                        break

                    iterations += 1
                    duration = wait_time * iterations
                    print("It is not possible to calculate the time of finalization. (at least: {})".format(daysHoursMinutes(duration)))
                    print("Proceed? [Y/n]")

                    try:
                        activate_input = read(values=["y", "Y", "n", "N", ""])
                    except KeyboardInterrupt:
                        iterations = 1
                        break

                    if activate_input.lower() == "n":
                        iterations = 1
                        banner()
                        continue
                break
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI activate the miracle {} {:d} times\n".format(
        island["wonderName"], iterations
    )
    setInfoSignal(session, info)
    try:
        do_it(session, island, iterations)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def wait_for_miracle(session, island):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict
    """
    while True:
        params = {
            "view": "temple",
            "cityId": island["ciudad"]["id"],
            "position": island["ciudad"]["pos"],
            "backgroundView": "city",
            "currentCityId": island["ciudad"]["id"],
            "actionRequest": actionRequest,
            "ajax": "1",
        }
        temple_response = session.post(params=params)
        temple_response = json.loads(temple_response, strict=False)
        temple_response = temple_response[2][1]

        for elem in temple_response:
            if "countdown" in temple_response[elem]:
                enddate = temple_response[elem]["countdown"]["enddate"]
                currentdate = temple_response[elem]["countdown"]["currentdate"]
                wait_time = enddate - currentdate
                next_activation_time = time.time() + wait_time
                session.setStatus(
                    f"Miracle {island['wonderName']} is activated. Available at: {getDateTime(next_activation_time)}"
                )
                break
        else:
            available = (
                temple_response["js_WonderViewButton"]["buttonState"] == "enabled"
            )
            if available:
                return
            else:
                wait_time = 60

        msg = "I wait {:d} seconds to activate the miracle {}".format(
            wait_time, island["wonderName"]
        )
        sendToBotDebug(session, msg, debugON_activateMiracle)
        wait(wait_time + 5)


def do_it(session, island, iterations):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict
    iterations : int
    """
    iterations_left = iterations
    session.setStatus(f"Waiting to activate {island['wonderName']}...")
    for i in range(iterations):

        wait_for_miracle(session, island)

        response = activateMiracleHttpCall(session, island)

        if response[1][1][0] == "error":
            msg = "The miracle {} could not be activated.".format(
                island["wonderName"]
            )
            sendToBot(session, msg)
            return
        iterations_left -= 1
        session.setStatus(
            f"Activated {island['wonderName']} @{getDateTime()}, iterations left: {iterations_left}"
        )
        msg = "Miracle {} successfully activated".format(island["wonderName"])
        sendToBotDebug(session, msg, debugON_activateMiracle)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import os
import subprocess

import psutil

from ikabot.config import *
from ikabot.helpers.signals import deactivate_sigint
from ikabot.helpers.varios import normalizeDicts


def set_child_mode(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    session.padre = False
    deactivate_sigint()


def run(command):
    ret = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ).stdout.read()
    try:
        return ret.decode("utf-8").strip()
    except Exception:
        return ret


def updateProcessList(session, programprocesslist=[]):
    """This function will return data about all the active ikabot processes. If it is passed the ``programprocesslist`` argument, it will write new processes from that list to the .ikabot file
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    programprocesslist : list[dict]
        a list of dictionaries containing relevant data about a running ikabot process ('pid', 'proxies' and 'action')

    Returns
    -------
    runningIkabotProcessList : list[dict]
        a list of dictionaries containing relevant data about a running ikabot process ('pid', 'proxies' and 'action')
    """
    # read from file
    sessionData = session.getSessionData()
    try:
        fileList = sessionData["processList"]
    except KeyError:
        fileList = []

    # check it's still running
    runningIkabotProcessList = []
    ika_process = psutil.Process(pid=os.getpid()).name()
    for process in fileList:
        try:
            proc = psutil.Process(pid=process["pid"])
        except psutil.NoSuchProcess:
            continue

        # windows doesn't support the status method
        isAlive = True if isWindows else proc.status() != "zombie"

        if proc.name() == ika_process and isAlive:
            runningIkabotProcessList.append(process)

    # add new to the list and write to file only if it's given
    for process in programprocesslist:
        if process not in runningIkabotProcessList:
            runningIkabotProcessList.append(process)

    # check if all proceses have new status field
    if len([p for p in runningIkabotProcessList if "status" not in p]) == len(
        runningIkabotProcessList
    ) and len(runningIkabotProcessList):
        runningIkabotProcessList[0]["status"] = "running"

    # write to file
    sessionData["processList"] = runningIkabotProcessList
    session.setSessionData(sessionData)

    # normalize process list (all processes must have properties pid, action, date and status)
    normalized_processes = normalizeDicts(runningIkabotProcessList)
    # remove dupes by pid
    return list({d["pid"]: d for d in normalized_processes}.values())



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import executeRoutes
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import addThousandSeparator


def sendResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        print("What type of ships do you want to use? (Default: Trade ships)")
        print("(1) Trade ships")
        print("(2) Freighters")
        shiptype = read(min=1, max=2, digit=True, empty=True)
        if shiptype == '':
            shiptype = 1
        if shiptype == 1:
            useFreighters = False
        elif shiptype == 2:
            useFreighters = True
        routes = []
        while True:

            banner()
            print("Origin city:")
            try:
                cityO = chooseCity(session)
            except KeyboardInterrupt:
                if routes:
                    print("Send shipment? [Y/n]")
                    rta = read(values=["y", "Y", "n", "N", ""])
                    if rta.lower() != "n":
                        break
                event.set()
                return

            banner()
            print("Destination city")
            cityD = chooseCity(session, foreign=True)
            idIsland = cityD["islandId"]

            if cityO["id"] == cityD["id"]:
                continue

            resources_left = cityO["availableResources"]
            for route in routes:
                (origin_city, destination_city, __, *toSend) = route
                if origin_city["id"] == cityO["id"]:
                    for i in range(len(materials_names)):
                        resources_left[i] -= toSend[i]

                # the destination city might be from another player
                if cityD["isOwnCity"] and destination_city["id"] == cityD["id"]:
                    for i in range(len(materials_names)):
                        cityD["freeSpaceForResources"][i] -= toSend[i]

            banner()
            # the destination city might be from another player
            if cityD["isOwnCity"]:
                msg = ""
                for i in range(len(materials_names)):
                    if resources_left[i] > cityD["freeSpaceForResources"][i]:
                        msg += "{} more {}\n".format(
                            addThousandSeparator(cityD["freeSpaceForResources"][i]),
                            materials_names[i].lower(),
                        )

                if len(msg) > 0:
                    print("You can store just:\n{}".format(msg))

            print("Available:")
            for i in range(len(materials_names)):
                print(
                    "{}:{} ".format(
                        materials_names[i], addThousandSeparator(resources_left[i])
                    ),
                    end="",
                )
            print("")

            print("Send:")
            try:
                max_name = max([len(material) for material in materials_names])
                send = []
                for i in range(len(materials_names)):
                    material_name = materials_names[i]
                    pad = " " * (max_name - len(material_name))
                    val = askForValue(
                        "{}{}:".format(pad, material_name), resources_left[i]
                    )
                    send.append(val)
            except KeyboardInterrupt:
                continue
            if sum(send) == 0:
                continue

            banner()
            print(
                "About to send from {} to {}".format(
                    cityO["cityName"], cityD["cityName"]
                )
            )
            for i in range(len(materials_names)):
                if send[i] > 0:
                    print(
                        "{}:{} ".format(
                            materials_names[i], addThousandSeparator(send[i])
                        ),
                        end="",
                    )
            print("")

            print("Proceed? [Y/n]")
            rta = read(values=["y", "Y", "n", "N", ""])
            if rta.lower() != "n":
                route = (cityO, cityD, idIsland, *send)
                routes.append(route)
                print("Create another shipment? [y/N]")
                rta = read(values=["y", "Y", "n", "N", ""])
                if rta.lower() != "y":
                    break
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nSend resources\n"

    setInfoSignal(session, info)
    try:
        executeRoutes(session, routes, useFreighters)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import copy
import json
import re
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import getAvailableResources
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *
from ikabot.helpers.varios import addThousandSeparator


def getBuildingInfo(session, city, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainTroops : bool

    Returns
    -------
    response : dict
    """
    view = "barracks" if trainTroops else "shipyard"
    params = {
        "view": view,
        "cityId": city["id"],
        "position": city["pos"],
        "backgroundView": "city",
        "currentCityId": city["id"],
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    data = session.post(params=params)
    return json.loads(data, strict=False)


def train(session, city, trainings, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainings : list[dict]
    trainTroops : bool
    """
    templateView = "barracks" if trainTroops else "shipyard"
    function = "buildUnits" if trainTroops else "buildShips"
    payload = {
        "action": "CityScreen",
        "function": function,
        "actionRequest": "REQUESTID",
        "cityId": city["id"],
        "position": city["pos"],
        "backgroundView": "city",
        "currentCityId": city["id"],
        "templateView": templateView,
        "ajax": "1",
    }
    for training in trainings:
        payload[training["unit_type_id"]] = training["train"]
    session.post(params=payload)


def waitForTraining(session, city, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainTroops : bool
    """
    data = getBuildingInfo(session, city, trainTroops)
    html = data[1][1][1]
    seconds = re.search(r"\'buildProgress\', (\d+),", html)
    if seconds:
        seconds = seconds.group(1)
        seconds = int(seconds) - data[0][1]["time"]
        wait(seconds + 5)


def planTrainings(session, city, trainings, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainings : list[list[dict]]
    trainTroops : bool
    """
    buildingPos = city["pos"]

    # trainings might be divided in multriple rounds
    while True:

        # total number of units to create
        total = sum([unit["cantidad"] for training in trainings for unit in training])
        if total == 0:
            return

        for training in trainings:
            waitForTraining(session, city, trainTroops)
            html = session.get(city_url + city["id"])
            city = getCity(html)
            city["pos"] = buildingPos

            resourcesAvailable = city["availableResources"].copy()
            resourcesAvailable.append(city["freeCitizens"])

            # for each unit type in training
            for unit in training:

                # calculate how many units can actually be trained based on the resources available
                unit["train"] = unit["cantidad"]

                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit["costs"]:
                        limiting = resourcesAvailable[i] // unit["costs"][material_name]
                        unit["train"] = min(unit["train"], limiting)

                if "citizens" in unit["costs"]:
                    limiting = (
                        resourcesAvailable[len(materials_names_english)]
                        // unit["costs"]["citizens"]
                    )
                    unit["train"] = min(unit["train"], limiting)

                # calculate the resources that will be left
                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit["costs"]:
                        resourcesAvailable[i] -= (
                            unit["costs"][material_name] * unit["train"]
                        )

                if "citizens" in unit["costs"]:
                    resourcesAvailable[len(materials_names_english)] -= (
                        unit["costs"]["citizens"] * unit["train"]
                    )

                unit["cantidad"] -= unit["train"]

            # amount of units that will be trained
            total = sum([unit["train"] for unit in training])
            if total == 0:
                msg = "It was not possible to finish the training due to lack of resources."
                
                sendToBot(session, msg)
                return

            train(session, city, training, trainTroops)


def generateArmyData(units_info):
    """
    Parameters
    ----------
    units_info : dict

    Returns
    -------
    units : list[dict]
    """
    i = 1
    units = []
    while "js_barracksSlider{:d}".format(i) in units_info:
        # {'identifier':'phalanx','unit_type_id':303,'costs':{'citizens':1,'wood':27,'sulfur':30,'upkeep':3,'completiontime':71.169695412658},'local_name':'Hoplita'}
        info = units_info["js_barracksSlider{:d}".format(i)]["slider"]["control_data"]
        info = json.loads(info, strict=False)
        units.append(info)
        i += 1
    return units


def trainArmy(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print("Do you want to train troops (1) or ships (2)?")
        rta = read(min=1, max=2)
        trainTroops = rta == 1
        banner()

        if trainTroops:
            print("In what city do you want to train the troops?")
        else:
            print("In what city do you want to train the fleet?")
        city = chooseCity(session)
        banner()

        lookfor = "barracks" if trainTroops else "shipyard"
        for i in range(len(city["position"])):
            if city["position"][i]["building"] == lookfor:
                city["pos"] = str(i)
                break
        else:
            if trainTroops:
                print("Barracks not built.")
            else:
                print("Shipyard not built.")
            enter()
            event.set()
            return

        data = getBuildingInfo(session, city, trainTroops)

        units_info = data[2][1]
        units = generateArmyData(units_info)

        maxSize = max([len(unit["local_name"]) for unit in units])

        tranings = []
        while True:
            units = generateArmyData(units_info)
            print("Train:")
            for unit in units:
                pad = " " * (maxSize - len(unit["local_name"]))
                amount = read(
                    msg="{}{}:".format(pad, unit["local_name"]), min=0, empty=True
                )
                if amount == "":
                    amount = 0
                unit["cantidad"] = amount

            # calculate costs
            cost = [0] * (len(materials_names_english) + 3)
            for unit in units:
                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit["costs"]:
                        cost[i] += unit["costs"][material_name] * unit["cantidad"]

                if "citizens" in unit["costs"]:
                    cost[len(materials_names_english) + 0] += (
                        unit["costs"]["citizens"] * unit["cantidad"]
                    )
                if "upkeep" in unit["costs"]:
                    cost[len(materials_names_english) + 1] += (
                        unit["costs"]["upkeep"] * unit["cantidad"]
                    )
                if "completiontime" in unit["costs"]:
                    cost[len(materials_names_english) + 2] += (
                        unit["costs"]["completiontime"] * unit["cantidad"]
                    )

            print("\nTotal cost:")
            for i in range(len(materials_names_english)):
                if cost[i] > 0:
                    print(
                        "{}: {}".format(
                            materials_names_english[i], addThousandSeparator(cost[i])
                        )
                    )
            if cost[len(materials_names_english) + 0] > 0:
                print(
                    "Citizens: {}".format(
                        addThousandSeparator(cost[len(materials_names_english) + 0])
                    )
                )
            if cost[len(materials_names_english) + 1] > 0:
                print(
                    "Maintenance: {}".format(
                        addThousandSeparator(cost[len(materials_names_english) + 1])
                    )
                )
            if cost[len(materials_names_english) + 2] > 0:
                print(
                    "Duration: {}".format(
                        daysHoursMinutes(int(cost[len(materials_names_english) + 2]))
                    )
                )

            print("\nProceed? [Y/n]")
            rta = read(values=["y", "Y", "n", "N", ""])
            if rta.lower() == "n":
                event.set()
                return

            tranings.append(units)

            if trainTroops:
                print("\nDo you want to train more troops when you finish? [y/N]")
            else:
                print("\nDo you want to train more fleets when you finish? [y/N]")
            rta = read(values=["y", "Y", "n", "N", ""])
            if rta.lower() == "y":
                banner()
                if trainTroops:
                    print("Train new troops (1) or repeat (2)?")
                else:
                    print("Train new fleets (1) or repeat (2)?")
                rta = read(min=1, max=2)
                if rta == 1:
                    continue
                else:
                    print("\nRepeat how many times?")
                    countRepeat = read(min=1, default=0)
                    break
            else:
                countRepeat = 0
                break

        print("Do you want to replicate the training to other cities? (y/N)")
        replicate = read(values=["y", "Y", "n", "N", ""])
        if replicate.lower() == "y":
            cityTrainings = []
            ids, cities = getIdsOfCities(session)

            print("(0) Back")
            print("(1) All the wine cities")
            print("(2) All the marble cities")
            print("(3) All the cristal cities")
            print("(4) All the sulfur cities")
            print("(5) Choose City")
            print("(6) All City")

            selected = read(min=0, max=6, digit=True)
            if selected == 0:
                event.set()
                return
            elif selected in [1, 2, 3, 4]:
                cityTrainings = filterCitiesByResource(
                    cities, resourceMapping[selected], cityTrainings
                )
            elif selected == 5:
                city = []
                while True:
                    city = chooseCity(session)
                    if city["id"] in cityTrainings:
                        print("\nYou have already selected this city!")
                        continue
                    cityTrainings.append(city["id"])
                    print("\nDo you want to add another city? [y/N]")
                    rta = read(values=["y", "Y", "n", "N", ""])
                    if rta.lower() == "y":
                        continue
                    else:
                        break
            elif selected == 6:
                ids, cities = getIdsOfCities(session)
                for cityId in ids:
                    cityTrainings.append(cityId)

        # calculate if the city has enough resources
        resourcesAvailable = city["availableResources"].copy()
        resourcesAvailable.append(city["freeCitizens"])

        for training in tranings:
            for unit in training:
                if unit["cantidad"] != 0:
                    for i in range(len(materials_names_english)):
                        material_name = materials_names_english[i].lower()
                        if material_name in unit["costs"]:
                            resourcesAvailable[i] -= (
                                unit["costs"][material_name] * unit["cantidad"]
                            )

                    if "citizens" in unit["costs"]:
                        resourcesAvailable[len(materials_names_english)] -= (
                            unit["costs"]["citizens"] * unit["cantidad"]
                        )

        not_enough = [elem for elem in resourcesAvailable if elem < 0] != []

        if not_enough:
            print("\nThere are not enough resources:")
            for i in range(len(materials_names_english)):
                if resourcesAvailable[i] < 0:
                    print(
                        "{}:{}".format(
                            materials_names[i],
                            addThousandSeparator(resourcesAvailable[i] * -1),
                        )
                    )

            if resourcesAvailable[len(materials_names_english)] < 0:
                print(
                    "Citizens:{}".format(
                        addThousandSeparator(
                            resourcesAvailable[len(materials_names_english)] * -1
                        )
                    )
                )

            print("\nProceed anyway? [Y/n]")
            rta = read(values=["y", "Y", "n", "N", ""])
            if rta.lower() == "n":
                event.set()
                return

        if trainTroops:
            print("\nThe selected troops will be trained.")
        else:
            print("\nThe selected fleet will be trained.")
        enter()

        if replicate == "y":
            countRepeat = countRepeat + 1
            while countRepeat > 0:
                for cityId in cityTrainings:
                    html = session.get(city_url + str(cityId))
                    city = getCity(html)

                    lookfor = "barracks" if trainTroops else "shipyard"
                    try:
                        for i in range(len(city["position"])):
                            if city["position"][i]["building"] == lookfor:
                                city["pos"] = str(i)
                                tranings_copy = []
                                units_copy = copy.deepcopy(units)
                                tranings_copy.append(units_copy)
                                loop = asyncio.get_event_loop()
                                loop.run_until_complete(
                                    planTrainings(
                                        session, city, tranings_copy, trainTroops
                                    )
                                )
                                break
                    except Exception as e:
                        if trainTroops:
                            info = "\nI train troops in {}\n".format(
                                city["cityName"]
                            )
                        else:
                            info = "\nI train fleets in {}\n".format(
                                city["cityName"]
                            )
                        msg = "Error in:\n{}\nCause:\n{}".format(
                            info, traceback.format_exc()
                        )
                        sendToBot(session, msg)
                countRepeat = countRepeat - 1

        else:
            if trainTroops:
                info = "\nI train troops in {}\n".format(city["cityName"])
            else:
                info = "\nI train fleets in {}\n".format(city["cityName"])
            setInfoSignal(session, info)
            try:
                planTrainings(session, city, tranings, trainTroops)
            except Exception as e:
                msg = "Error in:\n{}\nCause:\n{}".format(
                    info, traceback.format_exc()
                )
                sendToBot(session, msg)
            finally:
                session.logout()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()


def filterCitiesByResource(cities, resource_id, cityTrainings):
    for cityId, cityData in cities.items():
        if cityData["tradegood"] == resource_id:
            cityTrainings.append(cityData["id"])
    return cityTrainings


resourceMapping = {
    1: "1",  # Wine
    2: "2",  # Marble
    3: "3",  # Cristal
    4: "4",  # Sulfur
}



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import time
import datetime
import traceback
from decimal import *
import json

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import getIdsOfCities
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import getWineConsumptionPerHour, getAvailableResources, getProductionPerHour
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import daysHoursMinutes
from ikabot.helpers.planRoutes import *

getcontext().prec = 30

def alertLowWine(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return
        banner()
        hours = read(
            msg=(
                "How many hours should be left until the wine runs out in a city so that it's alerted? : "
            ),
            min=1,
        )
        auto_transfer = read(msg=("Would you like to automatically transfer wine if necessary? (y/n) : ")).strip().lower()
        
        if auto_transfer in ["y", "yes"]:
            auto_transfer = True
            transfer_amount = read(msg=("How much wine should be sent automatically? : "), min=1)
        else:
            auto_transfer = False
            transfer_amount = 0
        print("It will be alerted when the wine runs out in less than {:d} hours in any city, and {:,d} wine will be transferred if necessary.".format(hours, transfer_amount))

        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = ("\nI alert if the wine runs out in less than {:d} hours\n".format(hours))
    setInfoSignal(session, info)
    try:
        do_it(session, hours, auto_transfer, transfer_amount)
    except Exception as e:
        msg = (f"Error in:\n{info}\nCause:\n{traceback.format_exc()}")
        sendToBot(session, msg)
    finally:
        session.logout()

def getMovementsFromHtml(session):
    """
    Extract fleet movements using the same approach as See movements.
    Parameters
    ----------
    session : ikabot.web.session.Session
    Returns
    -------
    list
        A list of movements.
    """
    html = session.get()
    cityId = re.search(r"currentCityId:\s*(\d+),", html).group(1)
    url = "view=militaryAdvisor&oldView=city&oldBackgroundView=city&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1".format(
        cityId, actionRequest
    )
    resp = session.post(url)
    resp = json.loads(resp, strict=False)
    movements = resp[1][1][2]["viewScriptParams"]["militaryAndFleetMovements"]
    return movements

def isWineTransportInProgress(session, destinationCityId):
    """
    Check if there is an ongoing wine transport to the specified city.
    Parameters
    ----------
    session : ikabot.web.session.Session
    destinationCityId : str
    Returns
    -------
    bool
    """
    movements = getMovementsFromHtml(session)
    for movement in movements:
        target = movement.get("target", {})
        resources = movement.get("resources", [])

        # Convert both IDs to the same type for comparison
        target_city_id = str(target.get("cityId"))
        destination_city_id_str = str(destinationCityId)

        # Check if the target city matches and the transport includes wine
        if target_city_id == destination_city_id_str:
            wine_resource = next((res for res in resources if res["cssClass"] == "resource_icon wine"), None)
            if wine_resource:
                origin = movement.get("origin", {}).get("name", "Unknown")
                amount = wine_resource.get("amount", "Unknown")
                destination_name = movement.get("target", {}).get("name", "Unknown")
                return f"Active wine transport: {amount} wine from {origin} to {destination_name}."

    return None

def do_it(session, hours, auto_transfer, transfer_amount):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    hours : int
    auto_transfer : bool
    transfer_amount : int
    """
    was_alerted = {}
    message_log = []
    routes = []  # Store all routes for batch execution
    last_reset_time = datetime.datetime.now()
    
    while True:
        current_time = datetime.datetime.now()
        time_elapsed = (current_time - last_reset_time).total_seconds()
        if time_elapsed >= 12 * 60 * 60:  # 12 h to reset the alerted list
            was_alerted.clear()  # Reset all alerts
            last_reset_time = current_time  # Update the last time reseted the list
            
        ids, cities = getIdsOfCities(session)

        for cityId in cities:
            if cityId not in was_alerted:
                was_alerted[cityId] = False

        for cityId in cities:
            html = session.get(city_url + cityId)
            city = getCity(html)

            if "tavern" not in [building["building"] for building in city["position"]]:
                continue

            consumption_per_hour = getWineConsumptionPerHour(html)

            # Determine Wine Press reduction
            wine_press_level = 0
            for building in city["position"]:
                if building.get("building") == "vineyard":
                    wine_press_level = building.get("level", 0)
                    break

            # Apply reduction to wine consumption
            reduction_factor = Decimal(1 - (wine_press_level / 100))
            consumption_per_hour *= reduction_factor

            wine_available = Decimal(city["availableResources"][1])

            consumption_net = Decimal(consumption_per_hour)

            if consumption_net == 0:
                was_alerted[cityId] = False
                continue

            consumption_per_seg = consumption_net / Decimal(3600)
            seconds_left = wine_available / consumption_per_seg

            if seconds_left < hours * 60 * 60:
                if was_alerted[cityId] is False:
                    time_left = daysHoursMinutes(seconds_left)
                    message_log.append(f"In {city['name']} you have: {city['availableResources'][1]:,.0f} wine. Consumption: {consumption_per_hour:.2f} per hour.\nThe wine will run out in {time_left}")

                    if auto_transfer:
                        transport_status = isWineTransportInProgress(session, cityId)
                        if transport_status:
                            message_log.append(transport_status)
                            message_log.append(f"A wine transport is already in progress to {city['name']}. No additional transport initiated.")
                            was_alerted[cityId] = True
                            continue

                        donor_city_id = None
                        donor_city = None

                        # Find the donor city among wine-producing cities with the most wine available
                        max_wine_available = 0
                        for donor_id, donor in cities.items():
                            wood_prod, luxury_prod, tradegood = getProductionPerHour(session, donor_id)
                                
                            if tradegood != 1:  # Skip if not wine-producing
                                continue

                            donor_html = session.get(city_url + donor_id)
                            donor_info = getCity(donor_html)

                            if donor_id == cityId:
                                continue

                            # Check if the city has enough wine to send
                            wine_available = donor_info.get("availableResources", [0, 0, 0, 0, 0])[1]
                            if wine_available >= transfer_amount and wine_available > max_wine_available:
                                donor_city_id = donor_id
                                donor_city = donor
                                max_wine_available = wine_available

                        if donor_city_id:
                            routes.append((
                                donor_city,  # Origin
                                city,  # Destination
                                city["islandId"],  # Island ID
                                0,  # Wood
                                transfer_amount,  # Wine
                                0,  # Marble
                                0,  # Crystal
                                0,  # Sulfur
                            ))
                            
                            message_log.append(f"Will transfer {transfer_amount:,d} from {donor_city['name']}.")

                        else:
                            message_log.append(f"No city has sufficient wine to transfer to {city['name']}.")

                    was_alerted[cityId] = True
            else:
                was_alerted[cityId] = False

        if message_log:
            
            sendToBot(session, "\n".join(message_log))
            message_log.clear()

        if routes:
            executeRoutes(session, routes, useFreighters=False)
            routes.clear()
        time.sleep(60 * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-


import hashlib
import json
import math
import random
import re
import threading
import time
import traceback
from decimal import *

import requests
from functools import cache

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import getAvailableResources
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import *
from ikabot.web.session import normal_get

sendResources = True
expand = True
thread = None


def waitForConstruction(session, city_id, final_lvl):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int
    final_lvl : int

    Returns
    -------
    city : dict
    """
    while True:

        html = session.get(city_url + city_id)
        city = getCity(html)

        construction_buildings = [
            building for building in city["position"] if "completed" in building
        ]
        if len(construction_buildings) == 0:
            break

        construction_building = construction_buildings[0]
        construction_time = construction_building["completed"]

        current_time = int(time.time())
        final_time = int(construction_time)
        seconds_to_wait = final_time - current_time

        msg = "{}: I wait {:d} seconds so that {} gets to the level {:d}".format(
            city["cityName"],
            seconds_to_wait,
            construction_building["name"],
            construction_building["level"] + 1,
        )
        sendToBotDebug(session, msg, debugON_constructionList)
        session.setStatus(
            f"Waiting until {getDateTime(time.time()+seconds_to_wait+10)[8:]}, {construction_building['name']} {construction_building['level']} -> {construction_building['level']+1} in {city['name']}, final lvl: {final_lvl}"
        )
        wait(seconds_to_wait + 10)

    html = session.get(city_url + city_id)
    city = getCity(html)
    return city


def expandBuilding(session, cityId, building, waitForResources):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cityId : int
    building : dict
    waitForResources : bool
    """
    current_level = building["level"]
    if building["isBusy"]:
        current_level += 1
    levels_to_upgrade = building["upgradeTo"] - current_level
    position = building["position"]
    upgradeTo = building["upgradeTo"]
    time.sleep(
        random.randint(5, 15)
    )  # to avoid race conditions with sendResourcesNeeded

    for lv in range(levels_to_upgrade):
        city = waitForConstruction(session, cityId, upgradeTo)
        building = city["position"][position]

        if building["canUpgrade"] is False and waitForResources is True:
            while building["canUpgrade"] is False:
                time.sleep(60)
                seconds = getMinimumWaitingTime(session)
                html = session.get(city_url + cityId)
                city = getCity(html)
                building = city["position"][position]
                # if no ships are comming, exit no matter if the building can or can't upgrade
                if seconds == 0:
                    break
                wait(seconds + 5)

        if building["canUpgrade"] is False:
            msg = "City:{}\n".format(city["cityName"])
            msg += "Building:{}\n".format(building["name"])
            msg += "The building could not be completed due to lack of resources.\n"
            msg += "Missed {:d} levels".format(levels_to_upgrade - lv)
            sendToBot(session, msg)
            return

        url = "action=CityScreen&function=upgradeBuilding&actionRequest={}&cityId={}&position={:d}&level={}&activeTab=tabSendTransporter&backgroundView=city&currentCityId={}&templateView={}&ajax=1".format(
            actionRequest,
            cityId,
            position,
            building["level"],
            cityId,
            building["building"],
        )
        resp = session.post(url)
        html = session.get(city_url + cityId)
        city = getCity(html)
        building = city["position"][position]
        if building["isBusy"] is False:
            msg = "{}: The building {} was not extended".format(
                city["cityName"], building["name"]
            )
            sendToBot(session, msg)
            sendToBot(session, resp)
            return

        msg = "{}: The building {} is being extended to level {:d}.".format(
            city["cityName"], building["name"], building["level"] + 1
        )
        sendToBotDebug(session, msg, debugON_constructionList)

    msg = "{}: The building {} finished extending to level: {:d}.".format(
        city["cityName"], building["name"], building["level"] + 1
    )
    sendToBotDebug(session, msg, debugON_constructionList)


def getCostsReducers(city):
    """
    Parameters
    ----------
    city : dict

    Returns
    -------
    reducers_per_material_level : dict[int, int]
    """
    reducers_per_material = [0] * len(materials_names)
    assert len(reducers_per_material) == 5

    for building in city["position"]:
        if building["name"] == "empty":
            continue
        lv = building["level"]
        if building["building"] == "carpentering":
            reducers_per_material[0] = lv
        elif building["building"] == "vineyard":
            reducers_per_material[1] = lv
        elif building["building"] == "architect":
            reducers_per_material[2] = lv
        elif building["building"] == "optician":
            reducers_per_material[3] = lv
        elif building["building"] == "fireworker":
            reducers_per_material[4] = lv
    return reducers_per_material


def getResourcesNeeded(session, city, building, current_level, final_level):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    building : dict
    current_level : int
    final_level : int

    Returns
    -------
    costs_per_material : dict[int, int]
    """
    # get html with information about buildings
    building_detail_url = "view=buildingDetail&buildingId=0&helpId=1&backgroundView=city&currentCityId={}&templateView=ikipedia&actionRequest={}&ajax=1".format(
        city["id"], actionRequest
    )
    building_detail_response = session.post(building_detail_url)
    building_detail = json.loads(building_detail_response, strict=False)
    building_html = building_detail[1][1][1]

    # get html with information about buildings costs
    regex_building_detail = (
        r'<div class="(?:selected)? button_building '
        + re.escape(building["building"])
        + r'"\s*onmouseover="\$\(this\)\.addClass\(\'hover\'\);" onmouseout="\$\(this\)\.removeClass\(\'hover\'\);"\s*onclick="ajaxHandlerCall\(\'\?(.*?)\'\);'
    )
    match = re.search(regex_building_detail, building_html)
    building_costs_url = match.group(1)
    building_costs_url += "backgroundView=city&currentCityId={}&templateView=buildingDetail&actionRequest={}&ajax=1".format(
        city["id"], actionRequest
    )
    building_costs_response = session.post(building_costs_url)
    building_costs = json.loads(building_costs_response, strict=False)
    html_costs = building_costs[1][1][1]

    # if the user has all the resource saving studies, we save that in the session data (one less request)
    sessionData = session.getSessionData()
    if "reduccion_inv_max" in sessionData:
        costs_reduction = 14
    else:
        # get the studies
        url = "view=noViewChange&researchType=economy&backgroundView=city&currentCityId={}&templateView=researchAdvisor&actionRequest={}&ajax=1".format(
            city["id"], actionRequest
        )
        rta = session.post(url)
        rta = json.loads(rta, strict=False)
        studies = rta[2][1]["new_js_params"]
        studies = json.loads(studies, strict=False)
        studies = studies["currResearchType"]

        # look for resource saving studies
        costs_reduction = 0
        for study in studies:
            if studies[study]["liClass"] != "explored":
                continue
            link = studies[study]["aHref"]
            if "2020" in link:
                costs_reduction += 2
            elif "2060" in link:
                costs_reduction += 4
            elif "2100" in link:
                costs_reduction += 8

        # if the user has all the resource saving studies, save that in the session data
        if costs_reduction == 14:
            sessionData["reduccion_inv_max"] = True
            session.setSessionData(sessionData)

    # calculate cost reductions
    costs_reduction /= 100
    costs_reduction = 1 - costs_reduction

    # get buildings that reduce the cost of upgrades
    costs_reductions = getCostsReducers(city)

    # get the type of resources that this upgrade will cost (wood, marble, etc)
    resources_types = re.findall(
        r'<th class="costs"><img src="(.*?)\.png"/></th>', html_costs
    )[:-1]

    # get the actual cost of each upgrade
    matches = re.findall(
        r'<td class="level">\d+</td>(?:\s+<td class="costs">.*?</td>)+', html_costs
    )

    # calculate the cost of the entire upgrade, taking into account all the possible reductions
    final_costs = [0] * len(materials_names)
    levels_to_upgrade = 0
    for match in matches:
        lv = re.search(r'"level">(\d+)</td>', match).group(1)
        lv = int(lv)

        if lv <= current_level:
            continue
        if lv > final_level:
            break

        levels_to_upgrade += 1
        # get the costs for the current level
        costs = re.findall(r'<td class="costs"><div.*>([\d,\.\s\xa0]*)</div></div></td>', match)
        # delete blank spaces (\xa0) in costs
        costs = [value.replace('\xa0', '').replace(' ', '') for value in costs]

        for i in range(len(costs)):
            # get hash from CDN images to identify the resource type
            resource_type = checkhash("https:" + resources_types[i] + ".png")

            for j in range(len(materials_names_tec)):
                name = materials_names_tec[j]
                if resource_type == name:
                    resource_index = j
                    break

            # get the cost of the current resource type
            cost = costs[i]
            cost = cost.replace(",", "").replace(".", "")
            cost = 0 if cost == "" else int(cost)

            # calculate all the reductions
            real_cost = Decimal(cost)
            # investigation reduction
            original_cost = Decimal(real_cost) / Decimal(costs_reduction)
            # special building reduction
            real_cost -= Decimal(original_cost) * (
                Decimal(costs_reductions[resource_index]) / Decimal(100)
            )

            final_costs[resource_index] += math.ceil(real_cost)

    if levels_to_upgrade < final_level - current_level:
        print(
            "This building only allows you to expand {:d} more levels".format(
                levels_to_upgrade
            )
        )
        msg = "Expand {:d} levels? [Y/n]:".format(levels_to_upgrade)
        rta = read(msg=msg, values=["Y", "y", "N", "n", ""])
        if rta.lower() == "n":
            return [-1, -1, -1, -1, -1]

    return final_costs


def sendResourcesNeeded(session, destination_city_id, city_origins, missing_resources, useFreighters=False):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    destination_city_id : int
    city_origins : dict
    missing_resources : dict[int, int]
    """

    info = "\nTransport resources to upload building\n"

    try:
        routes = []
        html = session.get(city_url + destination_city_id)
        cityD = getCity(html)
        for i in range(len(materials_names)):
            missing = missing_resources[i]
            if missing <= 0:
                continue

            # send the resources from each origin city
            for cityOrigin in city_origins[i]:
                if missing == 0:
                    break

                available = cityOrigin["availableResources"][i]
                send = min(available, missing)
                missing -= send
                toSend = [0] * len(materials_names)
                toSend[i] = send
                route = (cityOrigin, cityD, cityD["islandId"], *toSend)
                routes.append(route)
        executeRoutes(session, routes, useFreighters)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
        # no s.logout() because this is a thread, not a process


def chooseResourceProviders(session, cities_ids, cities, city_id, resource, missing):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cities_ids : list[int]
    cities : dict[int, dict]
    city_id : int
    resource : int
    missing : int
    """
    global sendResources
    sendResources = True
    global expand
    expand = True

    banner()
    print("From what cities obtain {}?".format(materials_names[resource].lower()))

    tradegood_initials = [material_name[0] for material_name in materials_names]
    maxName = max(
        [len(cities[city]["name"]) for city in cities if cities[city]["id"] != city_id]
    )

    origin_cities = []
    total_available = 0
    for cityId in cities_ids:
        if cityId == city_id:
            continue

        html = session.get(city_url + cityId)
        city = getCity(html)

        available = city["availableResources"][resource]
        if available == 0:
            continue

        # ask the user it this city should provide resources
        tradegood_initial = tradegood_initials[int(cities[cityId]["tradegood"])]
        pad = " " * (maxName - len(cities[cityId]["name"]))
        is_producer = (int(cities[cityId]["tradegood"]) == int(resource))
        msg = "{}{} ({}): {} [{}]:".format(
            pad,
            cities[cityId]["name"],
            tradegood_initial,
            addThousandSeparator(available),
            ("y/N", "Y/n")[is_producer == True]
        )
        choice = read(msg=msg, values=["Y", "y", "N", "n", ""], default=("N", "Y")[is_producer == True])
        if choice.lower() == "n":
            continue

        # if so, save the city and calculate the total amount resources to send
        total_available += available
        origin_cities.append(city)
        # if we have enough resources, return
        if total_available >= missing:
            return origin_cities

    # if we reach this part, there are not enough resources to expand the building
    print("\nThere are not enough resources.")

    if len(origin_cities) > 0:
        print("\nSend the resources anyway? [Y/n]")
        choice = read(values=["y", "Y", "n", "N", ""])
        if choice.lower() == "n":
            sendResources = False

    print("\nTry to expand the building anyway? [y/N]")
    choice = read(values=["y", "Y", "n", "N", ""])
    if choice.lower() == "n" or choice == "":
        expand = False

    return origin_cities


def sendResourcesMenu(session, city_id, missing, useFreighters=False):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int
    missing : list[int, int]
    """
    global thread
    cities_ids, cities = getIdsOfCities(session)
    origins = {}
    # for each missing resource, choose providers
    for resource in range(len(missing)):
        if missing[resource] <= 0:
            continue

        origin_cities = chooseResourceProviders(
            session, cities_ids, cities, city_id, resource, missing[resource]
        )
        if sendResources is False and expand:
            print("\nThe building will be expanded if possible.")
            enter()
            return
        elif sendResources is False:
            return
        origins[resource] = origin_cities

    if expand:
        print(
            
                "\nThe resources will be sent and the building will be expanded if possible."
            
        )
    else:
        print("\nThe resources will be sent.")

    enter()

    # create a new thread to send the resources
    thread = threading.Thread(
        target=sendResourcesNeeded,
        args=(
            session,
            city_id,
            origins,
            missing,
            useFreighters,
        ),
    )
    thread.start()


def getBuildingsToExpand(session, cityId):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cityId : int

    Returns
    -------
    buildings : list of dict
    """
    html = session.get(city_url + cityId)
    city = getCity(html)

    banner()
    # show the buildings available to expand (ignore empty spaces)
    print("Which buildings do you want to expand? Separate numbers with commas (7, 1, 3, 5, ...)\n")
    print("(0)\t\texit")
    buildings = [
        building for building in city["position"] if building["name"] != "empty"
    ]
    for i in range(len(buildings)):
        building = buildings[i]

        level = building["level"]

        if building["isMaxLevel"] is True:
            color = bcolors.BLACK
        elif building["canUpgrade"] is True:
            color = bcolors.GREEN
        else:
            color = bcolors.RED
        if level < 10:
            level = " " + str(level)
        else:
            level = str(level)
        if building["isBusy"]:
            level = level + "+"
        print("({:d})\tlv:{}\t{}{}{}".format(i + 1, level, color, building["name"], bcolors.ENDC))

    selected_building_ids = read().split(",")
    selected_building_ids = [int(id.strip()) for id in selected_building_ids if id.strip().isdigit()]

    if len(selected_building_ids) == 0 or 0 in selected_building_ids:
        return None

    selected_buildings = []
    for building_id in selected_building_ids:
        building = buildings[building_id - 1]

        current_level = int(building["level"])
        # if the building is being expanded, add 1 level
        if building["isBusy"]:
            current_level += 1

        banner()
        print("building:{}".format(building["name"]))
        print("current level:{}".format(current_level))

        final_level = read(min=current_level, msg="increase to level:")
        building["upgradeTo"] = final_level
        selected_buildings.append(building)

    return selected_buildings

@cache
def checkhash(url):
    m = hashlib.md5()
    r = requests.get(url)
    for data in r.iter_content(8192):
        m.update(data)
        if m.hexdigest() == config.material_img_hash[0]:
            material = "wood"
        elif m.hexdigest() == config.material_img_hash[1]:
            material = "wine"
        elif m.hexdigest() == config.material_img_hash[2]:
            material = "marble"
        elif m.hexdigest() == config.material_img_hash[3]:
            material = "glass"
        elif m.hexdigest() == config.material_img_hash[4]:
            material = "sulfur"
        else:
            continue
    return material


def constructionList(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        global expand
        global sendResources
        expand = True
        sendResources = True

        banner()
        wait_resources = False
        print("In which city do you want to expand buildings?")
        city = chooseCity(session)
        cityId = city["id"]
        buildings = getBuildingsToExpand(session, cityId)
        if buildings is None or len(buildings) == 0:
            event.set()
            return

        # Sequentially upgrade each building
        for building in buildings:
            current_level = building["level"]
            if building["isBusy"]:
                current_level += 1
            final_level = building["upgradeTo"]

            # calculate the resources that are needed
            resourcesNeeded = getResourcesNeeded(
                session, city, building, current_level, final_level
            )
            if -1 in resourcesNeeded:
                event.set()
                return

            print("\nMaterials needed for {}:".format(building["name"]))
            for i, name in enumerate(materials_names):
                amount = resourcesNeeded[i]
                if amount == 0:
                    continue
                print("- {}: {}".format(name, addThousandSeparator(amount)))
            print("")

            # calculate the resources that are missing
            missing = [0] * len(materials_names)
            for i in range(len(materials_names)):
                if city["availableResources"][i] < resourcesNeeded[i]:
                    missing[i] = resourcesNeeded[i] - city["availableResources"][i]

            # show missing resources to the user
            if sum(missing) > 0:
                print("\nMissing:")
                for i in range(len(materials_names)):
                    if missing[i] == 0:
                        continue
                    name = materials_names[i].lower()
                    print("{} of {}".format(addThousandSeparator(missing[i]), name))
                print("")

                # if the user wants, send the resources from the selected cities
                print("Automatically transport resources? [Y/n]")
                rta = read(values=["y", "Y", "n", "N", ""])
                if rta.lower() == "n":
                    print("Proceed anyway? [Y/n]")
                    rta = read(values=["y", "Y", "n", "N", ""])
                    if rta.lower() == "n":
                        event.set()
                        return
                else:
                    print("What type of ships do you want to use? (Default: Trade ships)")
                    print("(1) Trade ships")
                    print("(2) Freighters")
                    shiptype = read(min=1, max=2, digit=True, empty=True)
                    if shiptype == '':
                        shiptype = 1
                    if shiptype == 1:
                        useFreighters = False
                    elif shiptype == 2:
                        useFreighters = True
                    wait_resources = True
                    sendResourcesMenu(session, cityId, missing, useFreighters)
            else:
                print("\nYou have enough materials")
                print("Proceed? [Y/n]")
                rta = read(values=["y", "Y", "n", "N", ""])
                if rta.lower() == "n":
                    event.set()
                    return
    except KeyboardInterrupt:
        event.set()
        return
    
    set_child_mode(session)
    event.set()

    info = "\nUpgrade building\n"
    info = info + "City: {}\nBuilding: {}. From {:d}, to {:d}".format(
        city["cityName"], building["name"], current_level, final_level
    )

    setInfoSignal(session, info)
    try:
        if expand:
            for building in buildings:
                expandBuilding(session, cityId, building, wait_resources)
        elif thread:
            thread.join()
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
from decimal import *

from ikabot.config import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.market import getGold
from ikabot.helpers.naval import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.resources import *
from ikabot.helpers.varios import *

getcontext().prec = 30


def getStatus(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        color_arr = [
            bcolors.ENDC,
            bcolors.HEADER,
            bcolors.STONE,
            bcolors.BLUE,
            bcolors.WARNING,
        ]

        (ids, __) = getIdsOfCities(session)
        total_resources = [0] * len(materials_names)
        total_production = [0] * len(materials_names)
        total_wine_consumption = 0
        city_population = {}
        housing_space = 0
        citizens = 0
        total_housing_space = 0
        total_citizens = 0
        available_ships = 0
        total_ships = 0
        for id in ids:
            session.get("view=city&cityId={}".format(id), noIndex=True)
            data = session.get("view=updateGlobalData&ajax=1", noIndex=True)
            json_data = json.loads(data, strict=False)
            json_data = json_data[0][1]["headerData"]
            if json_data["relatedCity"]["owncity"] != 1:
                continue
            wood = Decimal(json_data["resourceProduction"])
            good = Decimal(json_data["tradegoodProduction"])
            typeGood = int(json_data["producedTradegood"])
            total_production[0] += wood * 3600
            total_production[typeGood] += good * 3600
            total_wine_consumption += json_data["wineSpendings"]
            housing_space = int(json_data["currentResources"]["population"])
            city_population[id] = {
                "housing_space": housing_space
            }
            citizens = int(json_data["currentResources"]["citizens"])
            total_housing_space += housing_space
            total_citizens += citizens
            total_resources[0] += json_data["currentResources"]["resource"]
            total_resources[1] += json_data["currentResources"]["1"]
            total_resources[2] += json_data["currentResources"]["2"]
            total_resources[3] += json_data["currentResources"]["3"]
            total_resources[4] += json_data["currentResources"]["4"]
            available_ships = json_data["freeTransporters"]
            total_ships = json_data["maxTransporters"]
            total_gold = int(Decimal(json_data["gold"]))
            total_gold_production = int(
                Decimal(
                    json_data["scientistsUpkeep"]
                    + json_data["income"]
                    + json_data["upkeep"]
                )
            )
        print("Ships {:d}/{:d}".format(int(available_ships), int(total_ships)))
        print("\nTotal:")
        print("{:>10}".format(" "), end="|")
        for i in range(len(materials_names)):
            print("{:>12}".format(materials_names_english[i]), end="|")
        print()
        print("{:>10}".format("Available"), end="|")
        for i in range(len(materials_names)):
            print("{:>12}".format(addThousandSeparator(total_resources[i])), end="|")
        print()
        print("{:>10}".format("Production"), end="|")
        for i in range(len(materials_names)):
            print("{:>12}".format(addThousandSeparator(total_production[i])), end="|")
        print()
        print(
            "Housing Space: {}, Citizens: {}".format(
                addThousandSeparator(total_housing_space),
                addThousandSeparator(total_citizens),
            )
        )
        print(
            "Gold: {}, Gold production: {}".format(
                addThousandSeparator(total_gold),
                addThousandSeparator(total_gold_production),
            )
        )
        print(
            "Wine consumption: {}".format(addThousandSeparator(total_wine_consumption)),
            end="",
        )

        print("\nOf which city do you want to see the state?")
        city = chooseCity(session)
        city_id = city['id']
        banner()

        (wood, good, typeGood) = getProductionPerHour(session, city["id"])
        print(
            "\033[1m{}{}{}".format(
                color_arr[int(typeGood)], city["cityName"], color_arr[0]
            )
        )

        resources = city["availableResources"]
        storageCapacity = city["storageCapacity"]
        citizens = city["freeCitizens"]
        housing_space = city_population[city_id]["housing_space"]
        color_resources = []
        for i in range(len(materials_names)):
            if resources[i] == storageCapacity:
                color_resources.append(bcolors.RED)
            else:
                color_resources.append(bcolors.ENDC)
        print("Population:")
        print(
            "{}: {} {}: {}".format(
                "Housing space",
                addThousandSeparator(housing_space),
                "Citizens",
                addThousandSeparator(citizens),
            )
        )
        print("Storage: {}".format(addThousandSeparator(storageCapacity)))
        print("Resources:")
        for i in range(len(materials_names)):
            print(
                "{} {}{}{} ".format(
                    materials_names[i],
                    color_resources[i],
                    addThousandSeparator(resources[i]),
                    bcolors.ENDC,
                ),
                end="",
            )
        print("")

        print("Production:")
        print(
            "{}: {} {}: {}".format(
                materials_names[0],
                addThousandSeparator(wood),
                materials_names[typeGood],
                addThousandSeparator(good),
            )
        )

        hasTavern = "tavern" in [building["building"] for building in city["position"]]
        if hasTavern:
            consumption_per_hour = city["wineConsumptionPerHour"]
            if consumption_per_hour == 0:
                print(
                    "{}{}Does not consume wine!{}".format(
                        bcolors.RED, bcolors.BOLD, bcolors.ENDC
                    )
                )
            else:
                if typeGood == 1 and (good * 3600) > consumption_per_hour:
                    elapsed_time_run_out = "∞"
                else:
                    consumption_per_second = Decimal(consumption_per_hour) / Decimal(
                        3600
                    )
                    remaining_resources_to_consume = Decimal(resources[1]) / Decimal(
                        consumption_per_second
                    )
                    elapsed_time_run_out = daysHoursMinutes(
                        remaining_resources_to_consume
                    )
                print("There is wine for: {}".format(elapsed_time_run_out))

        for building in [
            building for building in city["position"] if building["name"] != "empty"
        ]:
            if building["isMaxLevel"] is True:
                color = bcolors.BLACK
            elif building["canUpgrade"] is True:
                color = bcolors.GREEN
            else:
                color = bcolors.RED

            level = building["level"]
            if level < 10:
                level = " " + str(level)
            else:
                level = str(level)
            if building["isBusy"] is True:
                level = level + "+"

            print(
                "lv:{}\t{}{}{}".format(level, color, building["name"], bcolors.ENDC)
            )

        enter()
        print("")
        event.set()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-


import json
import re

from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *



def constructBuilding(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print("City where to build:")
        city = chooseCity(session)
        banner()

        # list of free spaces in the selected city
        free_spaces = [
            buildings
            for buildings in city["position"]
            if buildings["building"] == "empty"
        ]

        # get a list of all the posible buildings that can be built
        buildings = []
        # different buildings can be built in different areas
        type_spaces = ["sea", "land", "shore", "wall"]
        for type_space in type_spaces:
            free_spaces_of_type = [
                free_space
                for free_space in free_spaces
                if free_space["type"] == type_space
            ]
            if len(free_spaces_of_type) > 0:
                # we take any space in the desired area
                free_space_of_type = free_spaces_of_type[0]
                params = {
                    "view": "buildingGround",
                    "cityId": city["id"],
                    "position": free_space_of_type["position"],
                    "backgroundView": "city",
                    "currentCityId": city["id"],
                    "actionRequest": actionRequest,
                    "ajax": "1",
                }
                buildings_response = session.post(params=params, noIndex=True)
                buildings_response = json.loads(buildings_response, strict=False)[1][1]
                if buildings_response == "":
                    continue
                html = buildings_response[1]
                matches = re.findall(
                    r'<li class="building (.+?)">\s*<div class="buildinginfo">\s*<div title="(.+?)"\s*class="buildingimg .+?"\s*onclick="ajaxHandlerCall\(\'.*?buildingId=(\d+)&',
                    html,
                )
                # add the buildings that can be built in this area
                for match in matches:
                    buildings.append(
                        {
                            "building": match[0],
                            "name": match[1],
                            "buildingId": match[2],
                            "type": type_space,
                        }
                    )

        if len(buildings) == 0:
            print("No building can be built.")
            enter()
            event.set()
            return

        # show list of buildings to the user
        print("What building do you want to build?\n")
        i = 0
        for building in buildings:
            i += 1
            print("({:d}) {}".format(i, building["name"]))
        selected_building_index = read(min=1, max=i)
        banner()

        # show posible positions for the selected building
        building = buildings[selected_building_index - 1]
        print("{}\n".format(building["name"]))
        options = [
            position_id
            for position_id in city["position"]
            if position_id["building"] == "empty"
            and position_id["type"] == building["type"]
        ]
        if len(options) == 1:
            option = options[0]
        else:
            print("In which position do you want to build?\n")
            i = 0
            for option in options:
                i += 1
                print("({:d}) {}".format(i, option["position"]))
            selected_building_index = read(min=1, max=i)
            option = options[selected_building_index - 1]
            banner()

        # build it
        params = {
            "action": "CityScreen",
            "function": "build",
            "cityId": city["id"],
            "position": option["position"],
            "building": building["buildingId"],
            "backgroundView": "city",
            "currentCityId": city["id"],
            "templateView": "buildingGround",
            "actionRequest": actionRequest,
            "ajax": "1",
        }
        buildings_response = session.post(params=params, noIndex=True)
        msg = json.loads(buildings_response, strict=False)[3][1][0]["text"]
        print(msg)
        enter()
        event.set()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re

import ikabot.config as config
from ikabot.helpers.process import run


def checkForUpdate():
    upgrade = run("python3 -m pip search ikabot")
    if "ikabot" not in upgrade:
        return

    upgrade = upgrade.split("\n")
    if len(upgrade) != 3:
        return

    upgrade = upgrade[2]
    match = re.search(r" +.*?: +(.*)", upgrade)
    if match is None:
        return

    new = match.group(1)
    config.update_msg = "[+] ikabot version {} is available\n".format(new)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import threading
import time
import traceback

from ikabot.function.vacationMode import activateVacationMode
from ikabot.helpers.botComm import *
from ikabot.helpers.gui import enter
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import daysHoursMinutes


def alertAttacks(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return

        banner()
        default = 20
        minutes = read(
            msg=
                "How often should I search for attacks?(min:3, default: {:d}): ".format(default),
            min=3,
            default=default,
        )
        # min_units = read(msg=_('Attacks with less than how many units should be ignored? (default: 0): '), digit=True, default=0)
        print("I will check for attacks every {:d} minutes".format(minutes))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI check for attacks every {:d} minutes\n".format(minutes)
    setInfoSignal(session, info)
    try:
        do_it(session, minutes)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def respondToAttack(session):
    """
    Parameters
    ---------
    session : ikabot.web.session.Session
    """

    # this allows the user to respond to an attack via telegram
    while True:
        time.sleep(60 * 3)
        responses = getUserResponse(session)
        for response in responses:
            # the response should be in the form of:
            # <pid>:<action number>
            rta = re.search(r"(\d+):?\s*(\d+)", response)
            if rta is None:
                continue

            pid = int(rta.group(1))
            action = int(rta.group(2))

            # if the pid doesn't match, we ignore it
            if pid != os.getpid():
                continue

            # currently just one action is supported
            if action == 1:
                # mv
                activateVacationMode(session)
            else:
                sendToBot(session, "Invalid command: {:d}".format(action))


def do_it(session, minutes):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    minutes : int
    """

    # this thread lets the user react to an attack once the alert is sent
    thread = threading.Thread(target=respondToAttack, args=(session,))
    thread.start()

    knownAttacks = []
    while True:
        ##Catch errors inside the function to not exit for any reason.
        currentAttacks = []
        try:
            # get the militaryMovements
            html = session.get()
            city_id = re.search(r"currentCityId:\s(\d+),", html).group(1)
            url = "view=militaryAdvisor&oldView=city&oldBackgroundView=city&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1".format(
                city_id, actionRequest
            )
            movements_response = session.post(url)
            postdata = json.loads(movements_response, strict=False)
            militaryMovements = postdata[1][1][2]["viewScriptParams"][
                "militaryAndFleetMovements"
            ]
            timeNow = int(postdata[0][1]["time"])

            for militaryMovement in [
                mov for mov in militaryMovements if mov["isHostile"]
            ]:
                event_id = militaryMovement["event"]["id"]
                currentAttacks.append(event_id)
                # if we already alerted this, do nothing
                if event_id not in knownAttacks:
                    knownAttacks.append(event_id)

                    # get information about the attack
                    missionText = militaryMovement["event"]["missionText"]
                    origin = militaryMovement["origin"]
                    target = militaryMovement["target"]
                    amountTroops = militaryMovement["army"]["amount"]
                    amountFleets = militaryMovement["fleet"]["amount"]
                    timeLeft = int(militaryMovement["eventTime"]) - timeNow

                    # send alert
                    msg = "-- ALERT --\n"
                    msg += missionText + "\n"
                    msg += "from the city {} of {}\n".format(
                        origin["name"], origin["avatarName"]
                    )
                    msg += "a {}\n".format(target["name"])
                    msg += "{} units\n".format(amountTroops)
                    msg += "{} fleet\n".format(amountFleets)
                    msg += "arrival in: {}\n".format(daysHoursMinutes(timeLeft))
                    msg += "If you want to put the account in vacation mode send:\n"
                    msg += "{:d}:1".format(os.getpid())
                    sendToBot(session, msg)

        except Exception as e:
            info = "\nI check for attacks every {:d} minutes\n".format(minutes)
            msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
            sendToBot(session, msg)

        # remove old attacks from knownAttacks
        for event_id in list(knownAttacks):
            if event_id not in currentAttacks:
                knownAttacks.remove(event_id)

        time.sleep(minutes * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import getpass
import os

from ikabot import config
from ikabot.config import *

def enter():
    """Wait for the user to press Enter"""
    try:
        if len(config.predetermined_input) > 0:
            return
    except Exception:
        pass
    if isWindows:
        input("\n[Enter]")  # TODO improve this
    else:
        getpass.getpass("\n[Enter]")


def clear():
    """Clears all text on the console"""
    if isWindows:
        os.system("cls")
    else:
        os.system("clear")


def banner():
    """Clears all text on the console and displays the Ikabot ASCII art banner"""
    clear()
    bner = f"""
    `7MMF'  `7MM                       `7MM\"""Yp,                 mm
      MM      MM                         MM    Yb                 MM
      MM      MM  ,MP'   ,6"Yb.          MM    dP    ,pW"Wq.    mmMMmm
      MM      MM ;Y     8)   MM          MM\"""bg.   6W'   `Wb     MM
      MM      MM;Mm      ,pm9MM          MM    `Y   8M     M8     MM
      MM      MM `Mb.   8M   MM          MM    ,9   YA.   ,A9     MM
    .JMML.  .JMML. YA.  `Moo9^Yo.      .JMMmmmd9     `Ybmd9'      `Mbmo
                                                            {IKABOT_VERSION_TAG}"""
    print("\n{}\n\n{}\n{}".format(bner, config.infoUser, config.update_msg))


def printChoiceList(list):
    """Prints the list with padded numbers next to each list entry.
    Parameters
    ----------
    list : list
        list to be printed
    """
    [
        print("{:>{pad}}) ".format(str(i + 1), pad=len(str(len(list)))) + str(item))
        for i, item in enumerate(list)
    ]


class bcolors:
    HEADER = "\033[95m"
    STONE = "\033[37m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    WARNING = "\033[93m"
    RED = "\033[91m"
    BLACK = "\033[90m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    DARK_RED = "\033[31m"
    DARK_BLUE = "\033[34m"
    DARK_GREEN = "\033[32m"
    CYAN = "\033[36m"
    YELLOW = "\033[33m"



#! /usr/bin/env python3
# -*- coding: utf-8 -*-


import sys
import time
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.gui import bcolors, enter
from ikabot.helpers.pedirInfo import chooseCity, getIdsOfCities
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import getDateTime, timeStringToSec, wait


earliest_wakeup_time = 24 * 60 * 60
wine_city = wood_city = luxury_city = favour_tasks = None


def loginDaily(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        global wine_city
        print("Choose the city where the daily login bonus wine will be sent:")
        wine_city = chooseCity(session)
        print("Do you want to automatically activate the cinetheatre bonus? (Y|N)")
        choice = read(values=["y", "Y", "n", "N"])
        if choice in ["y", "Y"]:

            # choose city for wood
            print("Choose the city where the wood bonus will be activated:")
            wood_city = chooseCity(session)

            # choose city for luxury resource
            print("Choose the city where the luxury resource bonus will be activated:")
            luxury_city = chooseCity(session)
        print("Do you want to collect the favour automatically? (Y|N)")
        choice = read(values=["y", "Y", "n", "N"])
        if choice in ["y", "Y"]:
            favour_tasks = [t for t in tasks]

            def modify_tasks():
                banner()
                print(
                    "Choose which daily tasks will be done/collected by Ikabot automatically."
                )
                print(
                    f"Tasks in {bcolors.BLUE}blue{bcolors.ENDC} WILL be done automatically, tasks in {bcolors.STONE}grey{bcolors.ENDC} WILL NOT be done."
                )
                print("Press [ENTER] or type in [Y] to confirm selection")
                for i, task in enumerate(tasks):
                    if task in favour_tasks:
                        print(i + 1, ") ", bcolors.BLUE, task, bcolors.ENDC)
                    else:
                        print(i + 1, ") ", bcolors.STONE, task, bcolors.ENDC)
                choice = read(
                    min=1,
                    max=len(tasks),
                    empty=True,
                    digit=True,
                    additionalValues=["y", "Y"],
                )
                if not choice or (not choice.isdigit() and choice.lower() == "y"):
                    return
                choice -= 1
                if list(tasks)[choice] in favour_tasks:
                    favour_tasks.remove(list(tasks)[choice])
                else:
                    favour_tasks.append(list(tasks)[choice])
                return modify_tasks()

            modify_tasks()

        print("I will do the thing.")
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI enter every day\n"
    setInfoSignal(session, info)
    try:
        do_it(
            session,
            wine_city=wine_city,
            wood_city=wood_city,
            luxury_city=luxury_city,
            favour_tasks=favour_tasks,
        )
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session, wine_city, wood_city, luxury_city, favour_tasks):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    message_sent = False
    while True:
        (ids, cities) = getIdsOfCities(session)
        global earliest_wakeup_time
        earliest_wakeup_time = 24 * 60 * 60
        # collect daily bonus wine
        html = session.post(city_url + str(wine_city["id"]))
        url = "action=AvatarAction&function=giveDailyActivityBonus&dailyActivityBonusCitySelect={0}&startPageShown=1&detectedDevice=1&autoLogin=on&cityId={0}&activeTab=multiTab2&backgroundView=city&currentCityId={0}&actionRequest={1}&ajax=1".format(
            wine_city["id"], actionRequest
        )
        session.post(url)

        def get_remaining_time_cinetheatre(features):
            ewt = 999999999
            for i, feature in enumerate(["Resource", "Tradegood", "Favour"]):
                if features[i] and f"js_nextPossible{feature}" in features[i]:
                    remaining_time_str = re.search(
                        rf'js_nextPossible{feature}\\">([\S\s]*?)<', features[i]
                    )
                    assert (
                        remaining_time_str
                    ), f"Could not get remaining time for {feature} cultural feature"
                    remaining_time_str = remaining_time_str.group(1).strip()
                    sec_to_wait = (
                        timeStringToSec(remaining_time_str) + 60
                    )  # add 60s because of rounding errors
                    ewt = sec_to_wait if sec_to_wait < ewt else ewt
            return ewt

        # collect wood cultural bonus
        if wood_city:
            html = session.post(city_url + str(wood_city["id"]))
            html = session.post(
                f"view=cinema&visit=1&currentCityId={wood_city['id']}&backgroundView=city&actionRequest={actionRequest}&ajax=1"
            )
            features = (
                html.split('id=\\"VideoRewards\\"')[1].split("ul>")[0].split("li>")
            )
            features = [f for f in features if "form" in f or "js_nextPossible" in f]
            if "js_nextPossibleResource" in features[0]:
                earliest_wakeup_time = (
                    get_remaining_time_cinetheatre(features)
                    if get_remaining_time_cinetheatre(features) < earliest_wakeup_time
                    else earliest_wakeup_time
                )
            else:
                videoId = re.search(
                    r'name=\\"videoId\\"\s*value=\\"(\d+)\\"', features[0]
                )
                assert videoId, "Could not find a match for videoId in html"
                videoId = videoId.group(1)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=requestBonus&bonusId=51&videoId={str(videoId)}&backgroundView=city&currentCityId={wood_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )
                session.setStatus("Waiting 55s to watch video for wood bonus")
                wait(55)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=watchVideo&videoId={str(videoId)}&backgroundView=city&currentCityId={wood_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )

        wait(1)

        # collect luxury good cultural bonus
        if luxury_city:
            html = session.post(city_url + str(luxury_city["id"]))
            html = session.post(
                f"view=cinema&visit=1&currentCityId={luxury_city['id']}&backgroundView=city&actionRequest={actionRequest}&ajax=1"
            )
            features = (
                html.split('id=\\"VideoRewards\\"')[1].split("ul>")[0].split("li>")
            )
            features = [f for f in features if "form" in f or "js_nextPossible" in f]
            if "js_nextPossibleTradegood" in features[1]:
                earliest_wakeup_time = (
                    get_remaining_time_cinetheatre(features)
                    if get_remaining_time_cinetheatre(features) < earliest_wakeup_time
                    else earliest_wakeup_time
                )
            else:
                videoId = re.search(
                    r'name=\\"videoId\\"\s*value=\\"(\d+)\\"', features[1]
                )
                assert videoId, "Could not find a match for videoId in html"
                videoId = videoId.group(1)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=requestBonus&bonusId=52&videoId={str(videoId)}&backgroundView=city&currentCityId={luxury_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )
                session.setStatus("Waiting 55s to watch video for luxury good bonus")
                wait(55)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=watchVideo&videoId={str(videoId)}&backgroundView=city&currentCityId={luxury_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )

        wait(1)

        # collect favour cultural bonus
        if wood_city:
            favour_cinetheater_city = wood_city
        if luxury_city:
            favour_cinetheater_city = luxury_city
        if favour_cinetheater_city:
            html = session.post(city_url + str(favour_cinetheater_city["id"]))
            html = session.post(
                f"view=cinema&visit=1&currentCityId={favour_cinetheater_city['id']}&backgroundView=city&actionRequest={actionRequest}&ajax=1"
            )
            features = (
                html.split('id=\\"VideoRewards\\"')[1].split("ul>")[0].split("li>")
            )
            features = [f for f in features if "form" in f or "js_nextPossible" in f]
            if "js_nextPossibleFavour" in features[2]:
                earliest_wakeup_time = (
                    get_remaining_time_cinetheatre(features)
                    if get_remaining_time_cinetheatre(features) < earliest_wakeup_time
                    else earliest_wakeup_time
                )
            else:
                videoId = re.search(
                    r'name=\\"videoId\\"\s*value=\\"(\d+)\\"', features[2]
                )
                assert videoId, "Could not find a match for videoId in html"
                videoId = videoId.group(1)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=requestBonus&bonusId=53&videoId={str(videoId)}&backgroundView=city&currentCityId={favour_cinetheater_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )
                session.setStatus(
                    "Waiting 55s to watch video for favour cultural bonus"
                )
                wait(55)
                session.post(
                    f"view=noViewChange&action=AdVideoRewardAction&function=watchVideo&videoId={str(videoId)}&backgroundView=city&currentCityId={favour_cinetheater_city['id']}&templateView=cinema&actionRequest={actionRequest}&ajax=1"
                )

        # get waiting times for all three cinetheatre features.

        wait(1)

        if wood_city or luxury_city or favour_cinetheater_city:
            html = session.post(city_url + str(wine_city["id"]))
            html = session.post(
                f"view=cinema&visit=1&currentCityId={wine_city['id']}&backgroundView=city&actionRequest={actionRequest}&ajax=1"
            )
            features = (
                html.split('id=\\"VideoRewards\\"')[1].split("ul>")[0].split("li>")
            )
            features = [f for f in features if "form" in f or "js_nextPossible" in f]
            earliest_wakeup_time = (
                get_remaining_time_cinetheatre(features)
                if get_remaining_time_cinetheatre(features) < earliest_wakeup_time
                else earliest_wakeup_time
            )

        wait(1)

        # do favour tasks in wine city
        for task in favour_tasks:
            html = session.post(city_url + str(wine_city["id"]))
            html = session.post(
                f"view=dailyTasks&backgroundView=city&currentCityId={wine_city['id']}&actionRequest={actionRequest}&ajax=1"
            )
            # check if favour is full (2500)
            match = re.search(r"currentFavor([\S\s]*?)(\d+)\s*<", html)
            assert match, "Can not obtain current favour amount"
            if match.group(2) == "2500":
                # send notification we are full on favour
                if not message_sent:
                    sendToBot(session, "Favour was not collected as you are full on it")
                    message_sent = (
                        True  # we don't want to spam the message, only once is enough
                    )
                break
            matches = re.findall("<tr([\S\s]*?)tr>", html)
            rows = [
                matches[1],
                matches[2],
                matches[4],
                matches[5],
                matches[7],
                matches[8],
                matches[10],
                matches[11],
            ]
            tasks[task](session, rows)
            match = re.search(
                r'"dailyTasksCountdown":{"countdown":{"enddate":(\d+),"currentdate":(\d+),"timeout_ajax',
                html,
            )
            assert match, "Can not get remaning daily tasks time"
            sec_remaining = (int(match.group(1)) - int(match.group(2))) - 600
            earliest_wakeup_time = (
                sec_remaining
                if sec_remaining < earliest_wakeup_time
                else earliest_wakeup_time
            )

        # find capital and get ambro bonus from fountain if it's active
        for id in ids:
            html = session.post(city_url + str(id))
            if 'class="fountain' in html:  # is capital
                if 'class="fountain_active' in html:  # foutain is active
                    url = "action=AmbrosiaFountainActions&function=collect&backgroundView=city&currentCityId={0}&templateView=ambrosiaFountain&actionRequest={1}&ajax=1".format(
                        id, actionRequest
                    )
                    session.post(url)
                break
            wait(1)
        earliest_wakeup_time = abs(
            earliest_wakeup_time
        )  # funnily enough, it's possible to have the earliest wakeup time be a negative number, this happens if there is less than 10 minutes remaning until tasks reset
        session.setStatus(
            f"Last activity @{getDateTime()}, next activity @{getDateTime(time.time()+earliest_wakeup_time)}"
        )
        wait(earliest_wakeup_time, 20)


def is_collectable(row):
    progress_left = (
        re.search(r"smallright progress details([\S\s]*?)>([\S\s]*?)<", row)
        .group(2)
        .strip()
        .replace(",", "")
    )
    progress_right = (
        re.search(r"left small progress details([\S\s]*?)>([\S\s]*?)<", row)
        .group(2)
        .strip()
        .replace(",", "")
    )
    return "textLineThrough" not in row and progress_left == progress_right


def collect_resource_favour(session, table):
    # literally just collect the first two tasks if they're done
    global wine_city
    passive1 = table[0]
    passive2 = table[1]

    if is_collectable(passive1):
        taskId = re.search(r'taskId=([\S\s]*?)\\"', passive1).group(1)
        session.post(
            f"action=CollectDailyTasksFavor&taskId={taskId}&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
        )
        wait(1)

    if is_collectable(passive2):
        taskId = re.search(r'taskId=([\S\s]*?)\\"', passive2).group(1)
        session.post(
            f"action=CollectDailyTasksFavor&taskId={taskId}&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
        )
        wait(1)


def look(session, table):
    # collect if collectable, otherwise look at specified view then collect
    global earliest_wakeup_time, wine_city
    for r in table:
        if "task_amount_28" in r:  # visit shop
            if is_collectable(r):
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=28&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
            else:
                session.post(
                    f"view=premium&linkType=2&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=28&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)

        if "task_amount_27" in r:  # open inventory
            if is_collectable(r):
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=27&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
            else:
                session.post(
                    f"view=inventory&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=27&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)

        if "task_amount_26" in r:  # look at highscore
            if is_collectable(r):
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=26&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
            else:
                session.post(
                    f"view=highscore&showMe=1&backgroundView=city&currentCityId={wine_city['id']}&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=26&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)


def capture_runs(session, table):
    # divide number of point by 115, take top integer, invoke autopirate for that many runs, collect bonus
    pass


def donate_wood(session, table):
    # find city with most wood and donate number of it to lux resource or wood production (randomly chosen), collect bonus
    pass


def stay_online_30_mins(session, table):
    # collect if possible, otherwise earliest wake time is 31 mins from now, then it will be collected in the next pass
    global earliest_wakeup_time, wine_city
    for r in table:
        if "task_amount_23" in r:  # stay online for 30 mins
            if is_collectable(r):
                session.post(
                    f"action=CollectDailyTasksFavor&taskId=23&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
                )
                wait(1)
            else:
                earliest_wakeup_time = (
                    31 * 60 if 31 * 60 < earliest_wakeup_time else earliest_wakeup_time
                )


def complete_tasks(session, table):
    # collect any still collectable tasks, this one should always run last
    global wine_city
    for row in table:
        if is_collectable(row):
            taskId = re.search(r'taskId=([\S\s]*?)\\"', row).group(1)
            session.post(
                f"action=CollectDailyTasksFavor&taskId={taskId}&ajax=1&backgroundView=city&currentCityId={wine_city['id']}&templateView=dailyTasks&actionRequest={actionRequest}&ajax=1"
            )
            wait(1)


tasks = {
    "Collect resource favour": collect_resource_favour,
    "Look at hightscore/shop/inventory": look,
    "Stay online for 30 mins": stay_online_30_mins,
    #'Conduct capture runs': capture_runs,
    #'Donate wood': donate_wood,
    "Complete 2/all tasks": complete_tasks,  # must be last
}



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import banner
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import executeRoutes
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import addThousandSeparator


def distributeResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print("What type of ships do you want to use? (Default: Trade ships)")
        print("(1) Trade ships")
        print("(2) Freighters")
        shiptype = read(min=1, max=2, digit=True, empty=True)
        if shiptype == '':
            shiptype = 1
        if shiptype == 1:
            useFreighters = False
        elif shiptype == 2:
            useFreighters = True
        print("What resource do you want to distribute?")
        print("(0) Exit")
        for i in range(len(materials_names)):
            print("({:d}) {}".format(i + 1, materials_names[i]))
        resource = read(min=0, max=5)
        if resource == 0:
            event.set()  # give main process control before exiting
            return
        resource -= 1

        if resource == 0:
            evenly = True
        else:
            print("\nHow do you want to distribute the resources?")
            print("1) From cities that produce them to cities that do not")
            print("2) Distribute them evenly among all cities")
            type_distribution = read(min=1, max=2)
            evenly = type_distribution == 2
        distribution_msg = 'Select the cities to participate in the distribution:'
        cities_ids, cities = ignoreCities(session, msg=distribution_msg)

        if evenly:
            routes = distribute_evenly(session, resource, cities_ids, cities)
        else:
            routes = distribute_unevenly(session, resource, cities_ids, cities)

        if routes is None:
            event.set()
            return

        banner()
        print("\nThe following shipments will be made:\n")
        for route in routes:
            print(
                "{} -> {} : {} {}".format(
                    route[0]["name"],
                    route[1]["name"],
                    route[resource + 3],
                    materials_names[resource],
                )
            )  # displays all routes to be executed in console

        print("\nProceed? [Y/n]")
        rta = read(values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            event.set()
            return

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()  # this is where we give back control to main process

    info = "\nDistribute {}\n".format(materials_names[resource])
    setInfoSignal(session, info)

    try:
        executeRoutes(session, routes, useFreighters)  # plan trips for all the routes
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)  # sends message to telegram bot
    finally:
        session.logout()


def distribute_evenly(session, resource_type, cities_ids, cities):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    resource_type : int
    """
    resourceTotal = 0

    originCities = {}
    destinationCities = {}
    allCities = {}
    for cityID in cities_ids:

        html = session.get(
            city_url + cityID
        )  # load html from the get request for that particular city
        city = getCity(html)  # convert the html to a city object

        resourceTotal += city["availableResources"][
            resource_type
        ]  # the cities resources are added to the total
        allCities[cityID] = city  # adds the city to all cities

    # if a city doesn't have enough storage to fit resourceAverage
    # ikabot will send enough resources to fill the store to the max
    # then, resourceAverage will be recalculated
    resourceAverage = resourceTotal // len(allCities)
    while True:

        len_prev = len(destinationCities)
        for cityID in allCities:
            if cityID in destinationCities:
                continue
            freeStorage = allCities[cityID]["freeSpaceForResources"][resource_type]
            storage = allCities[cityID]["storageCapacity"]
            if storage < resourceAverage:
                destinationCities[cityID] = freeStorage
                resourceTotal -= storage

        resourceAverage = resourceTotal // (len(allCities) - len(destinationCities))

        if len_prev == len(destinationCities):
            for cityID in allCities:
                if cityID in destinationCities:
                    continue
                if (
                    allCities[cityID]["availableResources"][resource_type]
                    > resourceAverage
                ):
                    originCities[cityID] = (
                        allCities[cityID]["availableResources"][resource_type]
                        - resourceAverage
                    )
                else:
                    destinationCities[cityID] = (
                        resourceAverage
                        - allCities[cityID]["availableResources"][resource_type]
                    )
            break

    originCities = {
        k: v
        for k, v in sorted(originCities.items(), key=lambda item: item[1], reverse=True)
    }  # sort origin cities in descending order
    destinationCities = {
        k: v for k, v in sorted(destinationCities.items(), key=lambda item: item[1])
    }  # sort destination cities in ascending order

    routes = []

    for originCityID in originCities:  # iterate through all origin city ids

        for (
            destinationCityID
        ) in destinationCities:  # iterate through all destination city ids
            if (
                originCities[originCityID] == 0
                or destinationCities[destinationCityID] == 0
            ):
                continue

            if (
                originCities[originCityID] > destinationCities[destinationCityID]
            ):  # if there's more resources above average in the origin city than resources below average in the destination city (origin city needs to have a surplus and destination city needs to have a deficit of resources for a route to be considered)
                toSend = destinationCities[
                    destinationCityID
                ]  # number of resources to send is the number of resources below average in destination city
            else:
                toSend = originCities[
                    originCityID
                ]  # send the amount of resources above average of the current origin city

            if toSend == 0:
                continue

            toSendArr = [0] * len(materials_names)
            toSendArr[resource_type] = toSend
            route = (
                allCities[originCityID],
                allCities[destinationCityID],
                allCities[destinationCityID]["islandId"],
                *toSendArr,
            )
            routes.append(route)

            # ROUTE BLOCK
            if originCities[originCityID] > destinationCities[destinationCityID]:
                originCities[originCityID] -= destinationCities[
                    destinationCityID
                ]  # remove the sent amount from the origin city's surplus
                destinationCities[destinationCityID] = (
                    0  # set the amount of resources below average in destination city to 0
                )
            else:
                destinationCities[destinationCityID] -= originCities[
                    originCityID
                ]  # remove the sent amount from the amount of resources below average in current destination city
                originCities[originCityID] = (
                    0  # set the amount of resources above average in origin city to 0
                )

    return routes


def distribute_unevenly(session, resource_type, cities_ids, cities):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    resource_type : int
    """
    total_available_resources_from_all_cities = 0
    origin_cities = {}
    destination_cities = {}
    for destination_city_id in cities_ids:
        is_city_mining_this_resource = (
            cities[destination_city_id]["tradegood"] == resource_type
        )
        if is_city_mining_this_resource:
            html = session.get(city_url + destination_city_id)
            city = getCity(html)
            if resource_type == 1:  # wine
                city["available_amount_of_resource"] = (
                    city["availableResources"][resource_type]
                    - city["wineConsumptionPerHour"]
                    - 1
                )
            else:
                city["available_amount_of_resource"] = city["availableResources"][
                    resource_type
                ]
            if city["available_amount_of_resource"] < 0:
                city["available_amount_of_resource"] = 0
            total_available_resources_from_all_cities += city[
                "available_amount_of_resource"
            ]
            origin_cities[destination_city_id] = city
        else:
            html = session.get(city_url + destination_city_id)
            city = getCity(html)
            city["free_storage_for_resource"] = city["freeSpaceForResources"][
                resource_type
            ]
            if city["free_storage_for_resource"] > 0:
                destination_cities[destination_city_id] = city

    if total_available_resources_from_all_cities <= 0:
        print("\nThere are no resources to send.")
        enter()
        return None
    if len(destination_cities) == 0:
        print("\nThere is no space available to send resources.")
        enter()
        return None

    remaining_resources_to_be_sent_to_each_city = (
        total_available_resources_from_all_cities // len(destination_cities)
    )
    free_storage_available_per_city = [
        destination_cities[city]["free_storage_for_resource"]
        for city in destination_cities
    ]
    total_free_storage_available_in_all_cities = sum(free_storage_available_per_city)
    remaining_resources_to_send = min(
        total_available_resources_from_all_cities,
        total_free_storage_available_in_all_cities,
    )
    toSend = {}

    while remaining_resources_to_send > 0:
        len_prev = len(toSend)
        for city_id in destination_cities:
            city = destination_cities[city_id]
            if (
                city_id not in toSend
                and city["free_storage_for_resource"]
                < remaining_resources_to_be_sent_to_each_city
            ):
                toSend[city_id] = city["free_storage_for_resource"]
                remaining_resources_to_send -= city["free_storage_for_resource"]

        if len(toSend) == len_prev:
            for city_id in destination_cities:
                if city_id not in toSend:
                    toSend[city_id] = remaining_resources_to_be_sent_to_each_city
            break

        free_storage_available_per_city = [
            destination_cities[city]["free_storage_for_resource"]
            for city in destination_cities
            if city not in toSend
        ]
        if len(free_storage_available_per_city) == 0:
            break
        total_free_storage_available_in_all_cities = sum(
            free_storage_available_per_city
        )
        remaining_resources_to_send = min(
            remaining_resources_to_send, total_free_storage_available_in_all_cities
        )
        remaining_resources_to_be_sent_to_each_city = (
            remaining_resources_to_send // len(free_storage_available_per_city)
        )

    routes = []
    for destination_city_id in destination_cities:
        destination_city = destination_cities[destination_city_id]
        island_id = destination_city["islandId"]
        missing_resources = toSend[destination_city_id]
        for origin_city_id in origin_cities:
            if missing_resources == 0:
                break

            origin_city = origin_cities[origin_city_id]
            resources_available_in_this_city = origin_city[
                "available_amount_of_resource"
            ]
            for route in routes:
                origin = route[0]
                resource = route[resource_type + 3]
                if origin["id"] == origin_city_id:
                    resources_available_in_this_city -= resource

            send_this_round = min(missing_resources, resources_available_in_this_city)
            available = destination_city["free_storage_for_resource"]
            if available == 0 or send_this_round == 0:
                continue

            if available < send_this_round:
                missing_resources = 0
                send_this_round = available
            else:
                missing_resources -= send_this_round

            toSendArr = [0] * len(materials_names)
            toSendArr[resource_type] = send_this_round
            route = (origin_city, destination_city, island_id, *toSendArr)

            routes.append(route)

    return routes



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import time
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getIsland
from ikabot.helpers.gui import enter
from ikabot.helpers.pedirInfo import getIslandsIds
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import getDateTime, wait


def searchForIslandSpaces(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return
        banner()
        print(
            "Do you want to search for spaces on your islands or a specific set of islands?"
        )
        print("(0) Exit")
        print("(1) Search all islands I have colonised")
        print("(2) Search a specific set of islands")
        choice = read(min=0, max=2)
        islandList = []
        if choice == 0:
            event.set()
            return
        elif choice == 2:
            banner()
            print(
                "Insert the coordinates of each island you want searched like so: X1:Y1, X2:Y2, X3:Y3..."
            )
            coords_string = read()
            coords_string = coords_string.replace(" ", "")
            coords = coords_string.split(",")
            for coord in coords:
                coord = "&xcoord=" + coord
                coord = coord.replace(":", "&ycoord=")
                html = session.get("view=island" + coord)
                island = getIsland(html)
                islandList.append(island["id"])
        else:
            pass

        banner()
        print(
            "How frequently should the islands be searched in minutes (minimum is 3)?"
        )
        time = read(min=3, digit=True)
        banner()
        print(
            "Do you wish to be notified when a fight breaks out and stops on a city on these islands? (Y|N)"
        )
        fights = read(values=["y", "Y", "n", "N"])
        banner()
        print("I will search for changes in the selected islands")
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI search for new spaces each hour\n"
    setInfoSignal(session, info)
    try:
        do_it(session, islandList, time, fights)
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session, islandList, time, fights):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    islandList : list[dict]
        A list containing island objects which should be searched, if an empty list is passed, all the user's colonised islands are searched
    time : int
        The time in minutes between two consecutive seraches
    fights : str
        String that can either be y or n. Indicates whether or not to scan for fight activity on islands
    """

    # this dict will contain all the cities from each island
    # as they where in last scan
    cities_before_per_island = {}

    while True:
        # this is done inside the loop because the user may colonize in a new island
        if islandList != []:
            islandsIds = islandList
        else:
            islandsIds = getIslandsIds(session)
        for islandId in islandsIds:
            html = session.get(island_url + islandId)
            island = getIsland(html)
            # cities in the current island
            cities_now = [
                city_space
                for city_space in island["cities"]
                if city_space["type"] != "empty"
            ]  # loads the islands non empty cities into ciudades

            # if we haven't scaned this island before,
            # save it and do nothing
            if islandId not in cities_before_per_island:
                cities_before_per_island[islandId] = cities_now.copy()
            else:
                cities_before = cities_before_per_island[islandId]

                # someone disappeared
                for city_before in cities_before:
                    if city_before["id"] not in [
                        city_now["id"] for city_now in cities_now
                    ]:
                        # we didn't find the city_before in the cities_now
                        msg = "The city {} of the player {} disappeared in {}:{}.\n{}-{} / S.M. lvl: {} / L.G. lvl: {}.".format(
                            city_before["name"],
                            city_before["Name"],
                            island["x"],
                            island["y"],
                            materials_names[int(island["tradegood"])],
                            island["wonderName"],
                            island["resourceLevel"],
                            island["tradegoodLevel"],
                        )
                        sendToBot(session, msg)

                    if fights.lower() == "y":
                        for city_now in cities_now:
                            if city_now["id"] == city_before["id"]:
                                if (
                                    "infos" in city_now
                                    and "infos" not in city_before
                                    and "armyAction" in city_now["infos"]
                                    and city_now["infos"]["armyAction"] == "fight"
                                ):
                                    msg = "A fight started in the city {} of the player {} on island {} {}:{} {}".format(
                                        city_before["name"],
                                        city_before["Name"],
                                        materials_names[int(island["tradegood"])],
                                        island["x"],
                                        island["y"],
                                        island["name"],
                                    )
                                    sendToBot(session, msg)
                                if (
                                    "infos" not in city_now
                                    and "infos" in city_before
                                    and "armyAction" in city_before["infos"]
                                    and city_before["infos"]["armyAction"] == "fight"
                                ):
                                    msg = "A fight stopped in the city {} of the player {} on island {} {}:{} {}".format(
                                        city_before["name"],
                                        city_before["Name"],
                                        materials_names[int(island["tradegood"])],
                                        island["x"],
                                        island["y"],
                                        island["name"],
                                    )
                                    sendToBot(session, msg)

                # someone colonised
                for city_now in cities_now:
                    if city_now["id"] not in [
                        city_before["id"] for city_before in cities_before
                    ]:
                        # we didn't find the city_now in the cities_before
                        msg = "Player {} created a new city {} in {} {}:{} {}".format(
                            city_now["Name"],
                            city_now["name"],
                            materials_names[int(island["tradegood"])],
                            island["x"],
                            island["y"],
                            island["name"],
                        )
                        sendToBot(session, msg)

                cities_before_per_island[islandId] = (
                    cities_now.copy()
                )  # update cities_before_per_island for the current island
        session.setStatus(
            f'Checked islands {str([int(i) for i in islandsIds]).replace(" ","")} @ {getDateTime()[-8:]}'
        )
        wait(time * 60)



import re

from ikabot.config import *
from ikabot.helpers.naval import getAvailableShips
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.pedirInfo import getShipCapacity

def getCityMilitaryData(session, city_id):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int

    returns
    -------
    data : dict
    """
    params = {
        "view": "cityMilitary",
        "activeTab": "tabUnits",
        "cityId": city_id,
        "currentTab": "tabUnits",
        "currentCityId": city_id,
        "templateView": "cityMilitary",
        "actionRequest": actionRequest,
        "ajax": "1",
    }
    data = session.post(params=params)
    data = json.loads(data, strict=False)
    return data[1][1][1]


def extractTooltipsAndValues(data):
    tooltips = re.findall(r'<div class="tooltip">(.*?)</div>', data)
    values = re.findall(r"<td>\s*([\d.,-]+)\s*</td>", data)
    return tooltips, values


def calculateTotals(tooltips, values):
    total_units = 0
    total_ships = 0

    desc_value_dict = {}

    for i, (tooltip, value) in enumerate(zip(tooltips, values)):
        value = value.replace(",", "")
        is_digit = value.isdigit()
        int_value = int(value) if is_digit else 0

        if value.isdigit() and int_value > 0:
            desc_value_dict.setdefault(tooltip, []).append(int_value)

            if i <= 14:
                total_units += int_value
            else:
                total_ships += int_value

    return desc_value_dict, total_units, total_ships


def getArmyAvailable(session, type_army, destination_city_id, origin_city_id, event):
    params = {
        "view": "deployment",
        "deploymentType": "army" if type_army else "fleet",
        "destinationCityId": destination_city_id,
        "backgroundView": "city",
        "currentCityId": origin_city_id,
        "actionRequest": actionRequest,
        "ajax": 1,
    }
    ship_capacity, freighter_capacity = getShipCapacity(session)

    data = session.post(params=params)
    amount_results = re.findall(r'<div class=\\"amount\\">(.*?)<\\/div>', data)

    if type_army:
        army_results = re.findall(
            r'name=\\"cargo_army_([^\\]+)_upkeep\\"\\n\s+value=\\"([^\\"]+)\\"', data
        )
        weight_total_ships = int(getAvailableShips(session)) * ship_capacity if type_army else 0
        weight_results = re.findall(r'<div class=\\"weight\\">(.*?)<\\/div>', data)
    else:
        army_results = re.findall(
            r'name=\\"cargo_fleet_([^\\]+)_upkeep\\"\\n\s+value=\\"([^\\"]+)\\"', data
        )

    army_available = {}
    weight_total_army = 0

    if army_results:
        for i, result in enumerate(army_results):
            army_code = result[0]
            army = "cargo_{}_{}_upkeep".format(
                "army" if type_army else "fleet", army_code
            )
            army_only = "cargo_{}_{}".format(
                "army" if type_army else "fleet", army_code
            )
            upkeep = result[1]
            quantity = amount_results[i]
            army_available[army] = upkeep
            army_available[army_only] = quantity
            if type_army and weight_results and int(weight_results[i]) > 0:
                weight_total_army += int(quantity) * int(weight_results[i])

        if type_army and weight_total_army > weight_total_ships:
            banner()
            print("Not enough ships to transport all the units!")
            enter()
            return None
        return army_available
    return None


def sendArmy(session, origin_city, destination_city, type_army, army_available):
    params = {
        "action": "transportOperations",
        "function": "deployArmy" if type_army else "deployFleet",
        "actionRequest": actionRequest,
        "islandId": destination_city["islandId"],
        "destinationCityId": destination_city["id"],
        "deploymentType": "army" if type_army else "fleet",
        "backgroundView": "city",
        "currentCityId": origin_city["id"],
        "templateView": "deployment",
        "ajax": 1,
    }

    for army in army_available:
        params[army] = army_available[army]

    session.post(params=params)


def stationArmy(session, event, stdin_fd, predetermined_input):
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    type_army = True
    try:
        banner()

        ids, cities = getIdsOfCities(session)
        army = {}
        print("Total:")
        print("{:>19}|{:>19}|{:>19}|".format("", "Units", "Ships"))

        for city_id in cities:
            city = cities[city_id]
            data = getCityMilitaryData(session, city["id"])
            desc, values = extractTooltipsAndValues(data)
            army, total_units, total_ships = calculateTotals(desc, values)

            print(
                "{:>19}|{:>19}|{:>19}|".format(
                    city["name"],
                    addThousandSeparator(total_units),
                    addThousandSeparator(total_ships),
                )
            )

        print()
        print("(0) Back")
        print("(1) Move troops")
        print("(2) Move ships")
        print("(3) Move all ground units to a city.")
        print("(4) Move all maritime units to a city.")
        print("(5) Move all units to a city.")

        selected = read(min=0, max=5, digit=True)
        if selected == 0:
            event.set()
            return
        elif selected in (1, 2):
            print("Origin city:")
            origin_city = chooseCity(session)
            print()
            print("Destination city:")
            destination_city = chooseCity(session)
            if origin_city["id"] == destination_city["id"]:
                banner()
                print("The city of origin and the destination city cannot be the same!")
                enter()
                event.set()
            else:
                type_army = selected == 1
                army_available = getArmyAvailable(
                    session, type_army, destination_city["id"], origin_city["id"], event
                )
                if army_available != None:
                    sendArmy(
                        session,
                        origin_city,
                        destination_city,
                        type_army,
                        army_available,
                    )
                    print("Army sent!")
                    enter()
                    event.set()
                else:
                    print()
                    print(
                        "No {} units available in {}.".format(
                            "ground" if type_army else "maritime", origin_city["name"]
                        )
                    )
                    enter()
                    event.set()
        elif selected in (3, 4, 5):
            print("Destination city:")
            destination_city = chooseCity(session)
            ids, cities = getIdsOfCities(session)

            if selected in (3, 5):
                type_army = True
                for city_id in cities:
                    if city_id != destination_city["id"]:
                        city = cities[city_id]
                        army_available = getArmyAvailable(
                            session,
                            type_army,
                            destination_city["id"],
                            city["id"],
                            event,
                        )
                        if army_available != None:
                            sendArmy(
                                session,
                                city,
                                destination_city,
                                type_army,
                                army_available,
                            )
                        else:
                            print(
                                "No ground units available in {}.".format(city["name"])
                            )
            if selected in (4, 5):
                type_army = False
                for city_id in cities:
                    if city_id != destination_city["id"]:
                        city = cities[city_id]
                        army_available = getArmyAvailable(
                            session,
                            type_army,
                            destination_city["id"],
                            city["id"],
                            event,
                        )
                        if army_available != None:
                            sendArmy(
                                session,
                                city,
                                destination_city,
                                type_army,
                                army_available,
                            )
                        else:
                            print(
                                "No maritime units available in {}.".format(
                                    city["name"]
                                )
                            )
            enter()
            event.set()
    except KeyboardInterrupt:
        event.set()
        return
